import { Realm } from "../experimental/realm.js";
import { docFromCommentDecorator, getIndexer } from "../lib/intrinsic/decorators.js";
import { $ } from "../typekit/index.js";
import { DuplicateTracker } from "../utils/duplicate-tracker.js";
import { MultiKeyMap, createRekeyableMap, isArray, mutate } from "../utils/misc.js";
import { createSymbol, getSymNode } from "./binder.js";
import { createChangeIdentifierCodeFix } from "./compiler-code-fixes/change-identifier.codefix.js";
import { createModelToObjectValueCodeFix, createTupleToArrayValueCodeFix, } from "./compiler-code-fixes/convert-to-value.codefix.js";
import { getDeprecationDetails, markDeprecated } from "./deprecation.js";
import { compilerAssert, ignoreDiagnostics } from "./diagnostics.js";
import { validateInheritanceDiscriminatedUnions } from "./helpers/discriminator-utils.js";
import { explainStringTemplateNotSerializable } from "./helpers/string-template-utils.js";
import { typeReferenceToString } from "./helpers/syntax-utils.js";
import { getEntityName, getTypeName } from "./helpers/type-name-utils.js";
import { marshallTypeForJS } from "./js-marshaller.js";
import { createDiagnostic } from "./messages.js";
import { Numeric } from "./numeric.js";
import { exprIsBareIdentifier, getFirstAncestor, getIdentifierContext, hasParseError, visitChildren, } from "./parser.js";
import { createTypeRelationChecker } from "./type-relation-checker.js";
import { getFullyQualifiedSymbolName, getParentTemplateNode, isArrayModelType, isErrorType, isNullType, isTemplateInstance, isType, isValue, } from "./type-utils.js";
import { IdentifierKind, ResolutionResultFlags, SyntaxKind, } from "./types.js";
/**
 * Maps type arguments to type instantiation.
 */
const TypeInstantiationMap = class extends MultiKeyMap {
};
export function createChecker(program, resolver) {
    const stats = {
        createdTypes: 0,
        finishedTypes: 0,
    };
    const stdTypes = {};
    const indeterminateEntities = new WeakMap();
    const docFromCommentForSym = new Map();
    const referenceSymCache = new WeakMap();
    const valueExactTypes = new WeakMap();
    let onCheckerDiagnostic = (x) => {
        program.reportDiagnostic(x);
    };
    const typePrototype = {};
    const globalNamespaceType = createGlobalNamespaceType();
    // Caches the deprecation test of nodes in the program
    const nodeDeprecationMap = new Map();
    const errorType = createType({ kind: "Intrinsic", name: "ErrorType" });
    const voidType = createType({ kind: "Intrinsic", name: "void" });
    const neverType = createType({ kind: "Intrinsic", name: "never" });
    const unknownType = createType({ kind: "Intrinsic", name: "unknown" });
    const nullType = createType({ kind: "Intrinsic", name: "null" });
    /**
     * Set keeping track of node pending type resolution.
     * Key is the SymId of a node. It can be retrieved with getNodeSymId(node)
     */
    const pendingResolutions = new PendingResolutions();
    const typespecNamespaceBinding = resolver.symbols.global.exports.get("TypeSpec");
    if (typespecNamespaceBinding) {
        initializeTypeSpecIntrinsics();
    }
    /**
     * Tracking the template parameters used or not.
     */
    const templateParameterUsageMap = new Map();
    const checker = {
        getTypeForNode,
        checkProgram,
        checkSourceFile,
        getLiteralType,
        getGlobalNamespaceType,
        cloneType,
        resolveRelatedSymbols,
        resolveCompletions,
        neverType,
        errorType,
        nullType,
        anyType: unknownType,
        voidType,
        typePrototype,
        createType,
        createAndFinishType,
        createLiteralType,
        finishType,
        isStdType,
        getStdType,
        resolveTypeReference,
        resolveTypeOrValueReference,
        getValueForNode,
        getTypeOrValueForNode,
        getValueExactType,
        getTemplateParameterUsageMap,
        isTypeAssignableTo: undefined,
        stats,
    };
    const relation = createTypeRelationChecker(program, checker);
    checker.isTypeAssignableTo = relation.isTypeAssignableTo;
    return checker;
    function getTemplateParameterUsageMap() {
        return templateParameterUsageMap;
    }
    function wrapInstantiationDiagnostic(diagnostic, templateMapper) {
        if (templateMapper === undefined || typeof diagnostic.target !== "object")
            return diagnostic;
        return {
            ...diagnostic,
            target: {
                node: diagnostic.target,
                templateMapper,
            },
        };
    }
    function reportCheckerDiagnostic(diagnostic, mapper) {
        onCheckerDiagnostic(wrapInstantiationDiagnostic(diagnostic, mapper));
    }
    function reportCheckerDiagnostics(diagnostics) {
        diagnostics.forEach((x) => reportCheckerDiagnostic(x));
    }
    function initializeTypeSpecIntrinsics() {
        // a utility function to log strings or numbers
        mutate(typespecNamespaceBinding.exports).set("log", {
            flags: 2048 /* SymbolFlags.Function */,
            name: "log",
            value(p, ...strs) {
                program.trace("log", strs.join(" "));
                return voidType;
            },
            declarations: [],
            node: undefined, // TODO: is this correct?
        });
        // Until we have an `unit` type for `null`
        mutate(resolver.symbols.null).type = nullType;
        getSymbolLinks(resolver.symbols.null).type = nullType;
    }
    function getStdType(name) {
        const type = stdTypes[name];
        if (type !== undefined) {
            return type;
        }
        const sym = typespecNamespaceBinding?.exports?.get(name);
        compilerAssert(sym, `Unexpected missing symbol to std type "${name}"`);
        if (sym.flags & 2 /* SymbolFlags.Model */) {
            checkModelStatement(sym.declarations[0], undefined);
        }
        else {
            checkScalar(sym.declarations[0], undefined);
        }
        const loadedType = stdTypes[name];
        compilerAssert(loadedType, `TypeSpec std type "${name}" should have been initalized before using array syntax.`);
        return loadedType;
    }
    /**
     * Create the link for the given type to the symbol links.
     * If currently instantiating a template it will link to the instantiations.
     * Else will link to the declaredType.
     * @param links Symbol link
     * @param type Type
     * @param mapper Type mapper if in an template instantiation
     */
    function linkType(links, type, mapper) {
        if (mapper === undefined) {
            links.declaredType = type;
            links.instantiations = new TypeInstantiationMap();
        }
        else if (links.instantiations) {
            links.instantiations.set(mapper.args, type);
        }
    }
    function linkMemberType(links, type, mapper) {
        if (mapper === undefined) {
            links.declaredType = type;
        }
    }
    /**
     * Check a member symbol.
     * @param sym Symbol binding a member node.
     * @param mapper Type mapper.
     * @returns Checked type for the given member symbol.
     */
    function checkMemberSym(sym, mapper) {
        const symbolLinks = getSymbolLinks(sym);
        const memberContainer = getTypeForNode(getSymNode(sym.parent), mapper);
        const type = symbolLinks.declaredType ?? symbolLinks.type;
        if (type) {
            return type;
        }
        else {
            return checkMember(getSymNode(sym), mapper, memberContainer);
        }
    }
    /**
     * Check a member node
     * @param node Member node to check
     * @param mapper Type mapper
     * @param containerType Member node container type(Interface, Model, Union, etc.)
     * @returns Checked member
     */
    function checkMember(node, mapper, containerType) {
        switch (node.kind) {
            case SyntaxKind.ModelProperty:
                return checkModelProperty(node, mapper);
            case SyntaxKind.EnumMember:
                return checkEnumMember(node, mapper, containerType);
            case SyntaxKind.OperationStatement:
                return checkOperation(node, mapper, containerType);
            case SyntaxKind.UnionVariant:
                return checkUnionVariant(node, mapper);
            case SyntaxKind.ScalarConstructor:
                return checkScalarConstructor(node, mapper, containerType);
        }
    }
    function getTypeForTypeOrIndeterminate(entity) {
        if (entity.entityKind === "Indeterminate") {
            return entity.type;
        }
        return entity;
    }
    function getTypeForNode(node, mapper) {
        const entity = checkNode(node, mapper);
        if (entity === null) {
            return errorType;
        }
        if (entity.entityKind === "Indeterminate") {
            return entity.type;
        }
        if (isValue(entity)) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "value-in-type",
                target: node,
            }));
            return errorType;
        }
        if (entity.kind === "TemplateParameter") {
            if (entity.constraint?.valueType) {
                // means this template constraint will accept values
                reportCheckerDiagnostic(createDiagnostic({
                    code: "value-in-type",
                    messageId: "referenceTemplate",
                    target: node,
                }));
            }
        }
        return entity;
    }
    function getValueForNode(node, mapper, constraint) {
        const initial = checkNode(node, mapper, constraint);
        if (initial === null) {
            return null;
        }
        let entity;
        if (initial.entityKind === "Indeterminate") {
            entity = getValueFromIndeterminate(initial.type, constraint, node);
        }
        else {
            entity = initial;
        }
        if (entity === null) {
            return null;
        }
        if (isValue(entity)) {
            return constraint ? inferScalarsFromConstraints(entity, constraint.type) : entity;
        }
        // If a template parameter that can be a value is used in a template declaration then we allow it but we return null because we don't have an actual value.
        if (entity.kind === "TemplateParameter" &&
            entity.constraint?.valueType &&
            entity.constraint.type === undefined &&
            mapper === undefined) {
            return createValue({
                entityKind: "Value",
                valueKind: "TemplateValue",
                type: entity.constraint.valueType,
            }, entity.constraint.valueType);
        }
        reportExpectedValue(node, entity);
        return null;
    }
    function reportExpectedValue(target, type) {
        if (type.kind === "Model" && type.name === "" && target.kind === SyntaxKind.ModelExpression) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "expect-value",
                messageId: "model",
                format: { name: getTypeName(type) },
                codefixes: [createModelToObjectValueCodeFix(target)],
                target,
            }));
        }
        else if (type.kind === "Tuple" && target.kind === SyntaxKind.TupleExpression) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "expect-value",
                messageId: "tuple",
                format: { name: getTypeName(type) },
                codefixes: [createTupleToArrayValueCodeFix(target)],
                target,
            }));
        }
        else {
            reportCheckerDiagnostic(createDiagnostic({
                code: "expect-value",
                format: { name: getTypeName(type) },
                target,
            }));
        }
    }
    /** In certain context for types that can also be value if the constraint allows it we try to use it as a value instead of a type. */
    function getValueFromIndeterminate(type, constraint, node) {
        switch (type.kind) {
            case "String":
            case "StringTemplate":
                return checkStringValue(type, constraint, node);
            case "Number":
                return checkNumericValue(type, constraint, node);
            case "Boolean":
                return checkBooleanValue(type, constraint, node);
            case "EnumMember":
                return checkEnumValue(type, constraint, node);
            case "UnionVariant":
                return getValueFromIndeterminate(type.type, constraint, node);
            case "Intrinsic":
                switch (type.name) {
                    case "null":
                        return checkNullValue(type, constraint, node);
                }
                return type;
            default:
                return type;
        }
    }
    /**
     * Try to intercept types used as values, either return null and emit a diagnostic with a codefix or return the type as is to be dealt with the normal way.
     */
    function interceptTypesUsedAsValue(type) {
        switch (type.kind) {
            case "Tuple":
                return interceptTupleUsedAsValue(type);
            case "Model":
                return interceptModelExpressionUsedAsValue(type);
            default:
                return type;
        }
    }
    // Legacy behavior to smooth transition to object values.
    function interceptModelExpressionUsedAsValue(model) {
        if (model.node?.kind !== SyntaxKind.ModelExpression) {
            return model; // we only want to convert model expressions
        }
        reportCheckerDiagnostic(createDiagnostic({
            code: "expect-value",
            codefixes: [createModelToObjectValueCodeFix(model.node)],
            messageId: "modelExpression",
            target: model.node,
        }));
        return null;
    }
    // Intercept tuple used as value and report diagnostic with a codefix
    function interceptTupleUsedAsValue(tuple) {
        if (tuple.node?.kind !== SyntaxKind.TupleExpression) {
            return tuple; // we won't convert dynamic tuples to array values
        }
        reportCheckerDiagnostic(createDiagnostic({
            code: "expect-value",
            codefixes: [createTupleToArrayValueCodeFix(tuple.node)],
            messageId: "tuple",
            target: tuple.node,
        }));
        return null;
    }
    /** If the constraint only expect a value we can try to intercept types passed to it and call `interceptTypeUsedAsValue` to get a better error */
    function shouldTryInterceptTypeUsedAsValue(constraint) {
        return Boolean(constraint?.valueType && !constraint.type);
    }
    /**
     * Gets a type or value depending on the node and current constraint.
     * For nodes that can be both type or values(e.g. string), the value will be returned if the constraint expect a value of that type even if the constrain also allows the type.
     * This means that if the constraint is `string | valueof string` passing `"abc"` will send the value `"abc"` and not the type `"abc"`.
     */
    function getTypeOrValueForNode(node, mapper, constraint) {
        const valueConstraint = extractValueOfConstraints(constraint);
        const entity = checkNode(node, mapper, valueConstraint);
        if (entity === null) {
            return entity;
        }
        else if (isType(entity)) {
            if (shouldTryInterceptTypeUsedAsValue(constraint?.constraint)) {
                return interceptTypesUsedAsValue(entity);
            }
            else {
                return entity;
            }
        }
        else if (isValue(entity)) {
            return entity;
        }
        compilerAssert(entity.entityKind === "Indeterminate", "Expected indeterminate entity");
        if (valueConstraint) {
            const valueDiagnostics = [];
            const oldDiagnosticHook = onCheckerDiagnostic;
            onCheckerDiagnostic = (x) => valueDiagnostics.push(x);
            const result = getValueFromIndeterminate(entity.type, valueConstraint, node);
            onCheckerDiagnostic = oldDiagnosticHook;
            if (result) {
                // If there were diagnostic reported but we still got a value this means that the value might be invalid.
                reportCheckerDiagnostics(valueDiagnostics);
                return result;
            }
        }
        return entity.type;
    }
    /** Extact the type constraint a value should match. */
    function extractValueOfConstraints(constraint) {
        if (constraint?.constraint.valueType) {
            return { kind: constraint.kind, type: constraint.constraint.valueType };
        }
        else {
            return undefined;
        }
    }
    /**
     * Gets a type, value or indeterminate depending on the node and current constraint.
     * For nodes that can be both type or values(e.g. string literals), an indeterminate entity will be returned.
     * It is the job of of the consumer to decide if it should be a type or a value depending on the context.
     */
    function checkNode(node, mapper, valueConstraint) {
        switch (node.kind) {
            case SyntaxKind.ModelExpression:
                return checkModel(node, mapper);
            case SyntaxKind.ModelStatement:
                return checkModel(node, mapper);
            case SyntaxKind.ModelProperty:
                return checkModelProperty(node, mapper);
            case SyntaxKind.ScalarStatement:
                return checkScalar(node, mapper);
            case SyntaxKind.AliasStatement:
                return checkAlias(node, mapper);
            case SyntaxKind.EnumStatement:
                return checkEnum(node, mapper);
            case SyntaxKind.EnumMember:
                return checkEnumMember(node, mapper);
            case SyntaxKind.InterfaceStatement:
                return checkInterface(node, mapper);
            case SyntaxKind.UnionStatement:
                return checkUnion(node, mapper);
            case SyntaxKind.UnionVariant:
                return checkUnionVariant(node, mapper);
            case SyntaxKind.NamespaceStatement:
            case SyntaxKind.JsNamespaceDeclaration:
                return checkNamespace(node);
            case SyntaxKind.OperationStatement:
                return checkOperation(node, mapper);
            case SyntaxKind.NumericLiteral:
                return checkNumericLiteral(node);
            case SyntaxKind.BooleanLiteral:
                return checkBooleanLiteral(node);
            case SyntaxKind.StringLiteral:
                return checkStringLiteral(node);
            case SyntaxKind.TupleExpression:
                return checkTupleExpression(node, mapper);
            case SyntaxKind.StringTemplateExpression:
                return checkStringTemplateExpresion(node, mapper);
            case SyntaxKind.ArrayExpression:
                return checkArrayExpression(node, mapper);
            case SyntaxKind.UnionExpression:
                return checkUnionExpression(node, mapper);
            case SyntaxKind.IntersectionExpression:
                return checkIntersectionExpression(node, mapper);
            case SyntaxKind.DecoratorDeclarationStatement:
                return checkDecoratorDeclaration(node, mapper);
            case SyntaxKind.FunctionDeclarationStatement:
                return checkFunctionDeclaration(node, mapper);
            case SyntaxKind.TypeReference:
                return checkTypeOrValueReference(node, mapper);
            case SyntaxKind.TemplateArgument:
                return checkTemplateArgument(node, mapper);
            case SyntaxKind.TemplateParameterDeclaration:
                return checkTemplateParameterDeclaration(node, mapper);
            case SyntaxKind.VoidKeyword:
                return voidType;
            case SyntaxKind.NeverKeyword:
                return neverType;
            case SyntaxKind.UnknownKeyword:
                return unknownType;
            case SyntaxKind.ObjectLiteral:
                return checkObjectValue(node, mapper, valueConstraint);
            case SyntaxKind.ArrayLiteral:
                return checkArrayValue(node, mapper, valueConstraint);
            case SyntaxKind.ConstStatement:
                return checkConst(node);
            case SyntaxKind.CallExpression:
                return checkCallExpression(node, mapper);
            case SyntaxKind.TypeOfExpression:
                return checkTypeOfExpression(node, mapper);
            case SyntaxKind.AugmentDecoratorStatement:
                return checkAugmentDecorator(node);
            case SyntaxKind.UsingStatement:
                return checkUsings(node);
            default:
                return errorType;
        }
    }
    /**
     * Return a fully qualified id of node
     */
    function getNodeSym(node) {
        const symbol = node.kind === SyntaxKind.OperationStatement &&
            node.parent?.kind === SyntaxKind.InterfaceStatement
            ? getSymbolForMember(node)
            : node.symbol;
        return symbol;
    }
    /**
     * Check if the given namespace is the standard library `TypeSpec` namespace.
     */
    function isTypeSpecNamespace(namespace) {
        return namespace.name === "TypeSpec" && namespace.namespace === globalNamespaceType;
    }
    /**
     * Check if the given type is defined right in the TypeSpec namespace.
     */
    function isInTypeSpecNamespace(type) {
        return Boolean(type.namespace && isTypeSpecNamespace(type.namespace));
    }
    function checkTemplateParameterDeclaration(node, mapper) {
        const parentNode = node.parent;
        const grandParentNode = parentNode.parent;
        const links = getSymbolLinks(node.symbol);
        if (!templateParameterUsageMap.has(node)) {
            templateParameterUsageMap.set(node, false);
        }
        if (pendingResolutions.has(getNodeSym(node), ResolutionKind.Constraint)) {
            if (mapper === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "circular-constraint",
                    format: { typeName: node.id.sv },
                    target: node.constraint,
                }));
            }
            return errorType;
        }
        let type = links.declaredType;
        if (type === undefined) {
            if (grandParentNode) {
                if (grandParentNode.locals?.has(node.id.sv)) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "shadow",
                        format: { name: node.id.sv },
                        target: node,
                    }));
                }
            }
            const index = parentNode.templateParameters.findIndex((v) => v === node);
            type = links.declaredType = createAndFinishType({
                kind: "TemplateParameter",
                node: node,
            });
            if (node.constraint) {
                pendingResolutions.start(getNodeSym(node), ResolutionKind.Constraint);
                type.constraint = getParamConstraintEntityForNode(node.constraint);
                pendingResolutions.finish(getNodeSym(node), ResolutionKind.Constraint);
            }
            if (node.default) {
                type.default = checkTemplateParameterDefault(node.default, parentNode.templateParameters, index, type.constraint);
            }
        }
        return mapper ? mapper.getMappedType(type) : type;
    }
    function getResolvedTypeParameterDefault(declaredType, node, mapper) {
        if (declaredType.default === undefined) {
            return undefined;
        }
        if ((isType(declaredType.default) && isErrorType(declaredType.default)) ||
            declaredType.default === null) {
            return declaredType.default;
        }
        return checkNode(node.default, mapper);
    }
    function checkTemplateParameterDefault(nodeDefault, templateParameters, index, constraint) {
        function visit(node) {
            const entity = checkNode(node);
            let hasError = false;
            if (entity !== null && "kind" in entity && entity.kind === "TemplateParameter") {
                for (let i = index; i < templateParameters.length; i++) {
                    if (entity.node?.symbol === templateParameters[i].symbol) {
                        reportCheckerDiagnostic(createDiagnostic({ code: "invalid-template-default", target: node }));
                        return undefined;
                    }
                }
                return entity;
            }
            visitChildren(node, (x) => {
                const visited = visit(x);
                if (visited === undefined) {
                    hasError = true;
                }
            });
            return hasError ? undefined : entity;
        }
        const type = visit(nodeDefault) ?? errorType;
        if (!("kind" in type && isErrorType(type)) && constraint) {
            checkTypeAssignable(type, constraint, nodeDefault);
        }
        return type;
    }
    /**
     * Check and resolve a type for the given type reference node.
     * @param node Node.
     * @param mapper Type mapper for template instantiation context.
     * @param instantiateTemplate If templated type should be instantiated if they haven't yet.
     * @returns Resolved type.
     */
    function checkTypeReference(node, mapper, instantiateTemplate = true) {
        const sym = resolveTypeReferenceSym(node, mapper);
        if (!sym) {
            return errorType;
        }
        const type = checkTypeReferenceSymbol(sym, node, mapper, instantiateTemplate);
        return type;
    }
    /**
     * Check and resolve a type for the given type reference node.
     * @param node Node.
     * @param mapper Type mapper for template instantiation context.
     * @param instantiateTemplate If templated type should be instantiated if they haven't yet.
     * @returns Resolved type.
     */
    function checkTypeOrValueReference(node, mapper, instantiateTemplate = true) {
        const sym = resolveTypeReferenceSym(node, mapper);
        if (!sym) {
            return errorType;
        }
        return checkTypeOrValueReferenceSymbol(sym, node, mapper, instantiateTemplate) ?? errorType;
    }
    function checkTemplateArgument(node, mapper) {
        return checkNode(node.argument, mapper);
    }
    function resolveTypeOrValueReference(node) {
        const oldDiagnosticHook = onCheckerDiagnostic;
        const diagnostics = [];
        onCheckerDiagnostic = (x) => diagnostics.push(x);
        const entity = checkTypeOrValueReference(node, undefined, false);
        onCheckerDiagnostic = oldDiagnosticHook;
        return [entity === errorType ? undefined : entity, diagnostics];
    }
    function resolveTypeReference(node) {
        const oldDiagnosticHook = onCheckerDiagnostic;
        const diagnostics = [];
        onCheckerDiagnostic = (x) => diagnostics.push(x);
        const type = checkTypeReference(node, undefined, false);
        onCheckerDiagnostic = oldDiagnosticHook;
        return [type === errorType ? undefined : type, diagnostics];
    }
    function copyDeprecation(sourceType, destType) {
        const deprecationDetails = getDeprecationDetails(program, sourceType);
        if (deprecationDetails) {
            markDeprecated(program, destType, deprecationDetails);
        }
    }
    function checkDeprecated(type, node, target) {
        if (node) {
            const deprecationDetails = getDeprecationDetails(program, node);
            if (deprecationDetails) {
                reportDeprecation(program, target, deprecationDetails.message, reportCheckerDiagnostic);
                return;
            }
        }
        const deprecationDetails = getDeprecationDetails(program, type);
        if (deprecationDetails) {
            reportDeprecation(program, target, deprecationDetails.message, reportCheckerDiagnostic);
        }
    }
    function isTypeReferenceContextDeprecated(node) {
        function checkDeprecatedNode(node) {
            // Perform a simple check if the parent node is deprecated.  We do this
            // out of band because `checkDirectives` usually gets called on the parent
            // type after child types have already been checked (including their
            // deprecations).
            if (!nodeDeprecationMap.has(node)) {
                nodeDeprecationMap.set(node, (node.directives ?? []).findIndex((d) => d.target.sv === "deprecated") >= 0);
            }
            return nodeDeprecationMap.get(node);
        }
        // Walk the parent hierarchy up to a node which might have a
        // deprecation which would mitigate the deprecation warning of the original
        // type reference. This is done to prevent multiple deprecation notices from
        // being raised when a parent context is already being deprecated.
        switch (node.kind) {
            case SyntaxKind.ModelStatement:
                return checkDeprecatedNode(node);
            case SyntaxKind.OperationStatement:
                return (checkDeprecatedNode(node) ||
                    (node.parent.kind === SyntaxKind.InterfaceStatement &&
                        isTypeReferenceContextDeprecated(node.parent)));
            case SyntaxKind.InterfaceStatement:
                return checkDeprecatedNode(node);
            case SyntaxKind.IntersectionExpression:
            case SyntaxKind.UnionExpression:
            case SyntaxKind.ModelProperty:
            case SyntaxKind.OperationSignatureDeclaration:
            case SyntaxKind.OperationSignatureReference:
                return isTypeReferenceContextDeprecated(node.parent);
            default:
                return false;
        }
    }
    function checkTemplateInstantiationArgs(node, args, decls, mapper, parentMapper) {
        const params = new Map();
        const positional = [];
        const initMap = new Map(decls.map((decl) => {
            const declaredType = checkTemplateParameterDeclaration(decl, undefined);
            positional.push(declaredType);
            params.set(decl.id.sv, declaredType);
            return [
                declaredType,
                {
                    decl,
                    checkArgument: null,
                },
            ];
        }));
        let named = false;
        for (const [arg, idx] of args.map((v, i) => [v, i])) {
            function deferredCheck() {
                return [arg, checkNode(arg.argument, mapper)];
            }
            if (arg.name) {
                named = true;
                const param = params.get(arg.name.sv);
                if (!param) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-template-args",
                        messageId: "unknownName",
                        format: {
                            name: arg.name.sv,
                        },
                        target: arg,
                    }));
                    continue;
                }
                if (initMap.get(param).checkArgument !== null) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-template-args",
                        messageId: "specifiedAgain",
                        format: {
                            name: arg.name.sv,
                        },
                        target: arg,
                    }));
                    continue;
                }
                initMap.get(param).checkArgument = deferredCheck;
            }
            else {
                if (named) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-template-args",
                        messageId: "positionalAfterNamed",
                        target: arg,
                    }));
                    // we just throw this arg away. any missing args will be filled with ErrorType
                }
                if (idx >= positional.length) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-template-args",
                        messageId: "tooMany",
                        target: node,
                    }));
                    continue;
                }
                const param = positional[idx];
                initMap.get(param).checkArgument ??= deferredCheck;
            }
        }
        const finalMap = initMap;
        const mapperParams = [];
        const mapperArgs = [];
        for (const [param, { decl, checkArgument: init }] of [...initMap]) {
            function commit(param, type) {
                finalMap.set(param, type);
                mapperParams.push(param);
                mapperArgs.push(type);
            }
            if (init === null) {
                const argumentMapper = createTypeMapper(mapperParams, mapperArgs, { node, mapper }, parentMapper);
                const defaultValue = getResolvedTypeParameterDefault(param, decl, argumentMapper);
                if (defaultValue) {
                    commit(param, defaultValue);
                }
                else {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-template-args",
                        messageId: "missing",
                        format: {
                            name: decl.id.sv,
                        },
                        target: node,
                    }));
                    // TODO-TIM check if we expose this below
                    commit(param, param.constraint?.type ?? unknownType);
                }
                continue;
            }
            const [argNode, type] = init();
            if (type === null) {
                commit(param, unknownType);
                continue;
            }
            if (param.constraint) {
                const constraint = param.constraint.type?.kind === "TemplateParameter"
                    ? finalMap.get(param.constraint.type)
                    : param.constraint;
                if (isType(type) && shouldTryInterceptTypeUsedAsValue(param.constraint)) {
                    const converted = interceptTypesUsedAsValue(type);
                    // If we manage to convert it means this might be convertable so we skip type checking.
                    // However we still return the original entity
                    if (converted !== type) {
                        commit(param, type);
                        continue;
                    }
                }
                if (param.constraint && !checkArgumentAssignable(type, constraint, argNode)) {
                    const effectiveType = param.constraint.type ?? unknownType;
                    commit(param, effectiveType);
                    continue;
                }
            }
            else if (isErrorType(type)) {
                // If we got an error type we don't want to keep passing it through so we reduce to unknown
                // Similar to the above where if the type is not assignable to the constraint we reduce to the constraint
                commit(param, unknownType);
                continue;
            }
            else if (isValue(type)) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "value-in-type",
                    messageId: "noTemplateConstraint",
                    target: argNode,
                }));
                commit(param, unknownType);
                continue;
            }
            commit(param, type);
        }
        return finalMap;
    }
    /**
     * Check and resolve the type for the given symbol + node.
     * @param sym Symbol
     * @param node Node
     * @param mapper Type mapper for template instantiation context.
     * @param instantiateTemplates If a templated type should be instantiated if not yet @default true
     * @returns resolved type.
     */
    function checkTypeReferenceSymbol(sym, node, mapper, instantiateTemplates = true) {
        const result = checkTypeOrValueReferenceSymbol(sym, node, mapper, instantiateTemplates);
        if (result === null || isValue(result)) {
            reportCheckerDiagnostic(createDiagnostic({ code: "value-in-type", target: node }));
            return errorType;
        }
        if (result.entityKind === "Indeterminate") {
            return result.type;
        }
        return result;
    }
    function checkTypeOrValueReferenceSymbol(sym, node, mapper, instantiateTemplates = true) {
        const entity = checkTypeOrValueReferenceSymbolWorker(sym, node, mapper, instantiateTemplates);
        if (entity !== null && isType(entity) && entity.kind === "TemplateParameter") {
            templateParameterUsageMap.set(entity.node, true);
        }
        return entity;
    }
    function checkTypeOrValueReferenceSymbolWorker(sym, node, mapper, instantiateTemplates = true) {
        if (sym.flags & 131072 /* SymbolFlags.Const */) {
            return getValueForNode(sym.declarations[0], mapper);
        }
        if (sym.flags & 512 /* SymbolFlags.Decorator */) {
            reportCheckerDiagnostic(createDiagnostic({ code: "invalid-type-ref", messageId: "decorator", target: sym }));
            return errorType;
        }
        if (sym.flags & 2048 /* SymbolFlags.Function */) {
            reportCheckerDiagnostic(createDiagnostic({ code: "invalid-type-ref", messageId: "function", target: sym }));
            return errorType;
        }
        const argumentNodes = node.kind === SyntaxKind.TypeReference ? node.arguments : [];
        const symbolLinks = getSymbolLinks(sym);
        let baseType;
        if (sym.flags & 1048576 /* SymbolFlags.Declaration */ &&
            sym.flags &
                (2 /* SymbolFlags.Model */ |
                    4 /* SymbolFlags.Scalar */ |
                    128 /* SymbolFlags.Alias */ |
                    32 /* SymbolFlags.Interface */ |
                    8 /* SymbolFlags.Operation */ |
                    64 /* SymbolFlags.Union */)) {
            const decl = sym.declarations[0];
            if (!isTemplatedNode(decl)) {
                if (argumentNodes.length > 0) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-template-args",
                        messageId: "notTemplate",
                        target: node,
                    }));
                }
                if (sym.flags & 4194304 /* SymbolFlags.LateBound */) {
                    compilerAssert(sym.type, "Expected late bound symbol to have type");
                    return sym.type;
                }
                else if (symbolLinks.declaredType) {
                    baseType = symbolLinks.declaredType;
                }
                else if (sym.flags & 65536 /* SymbolFlags.Member */) {
                    baseType = checkMemberSym(sym, mapper);
                }
                else {
                    baseType = checkDeclaredTypeOrIndeterminate(sym, decl, mapper);
                }
            }
            else {
                const declaredType = getOrCheckDeclaredType(sym, decl, mapper);
                const templateParameters = decl.templateParameters;
                const instantiation = checkTemplateInstantiationArgs(node, argumentNodes, templateParameters, mapper, declaredType.templateMapper);
                baseType = getOrInstantiateTemplate(decl, [...instantiation.keys()], [...instantiation.values()], { node, mapper }, declaredType.templateMapper, instantiateTemplates);
            }
        }
        else {
            const symNode = getSymNode(sym);
            // some other kind of reference
            if (argumentNodes.length > 0) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "invalid-template-args",
                    messageId: "notTemplate",
                    target: node,
                }));
            }
            if (sym.flags & 4194304 /* SymbolFlags.LateBound */) {
                compilerAssert(sym.type, `Expected late bound symbol to have type`);
                return sym.type;
            }
            else if (sym.flags & 1024 /* SymbolFlags.TemplateParameter */) {
                const mapped = checkTemplateParameterDeclaration(symNode, mapper);
                baseType = mapped;
            }
            else if (symbolLinks.type) {
                // Have a cached type for non-declarations
                baseType = symbolLinks.type;
            }
            else if (symbolLinks.declaredType) {
                baseType = symbolLinks.declaredType;
            }
            else {
                if (sym.flags & 65536 /* SymbolFlags.Member */) {
                    baseType = checkMemberSym(sym, mapper);
                }
                else {
                    // don't have a cached type for this symbol, so go grab it and cache it
                    baseType = getTypeForNode(symNode, mapper);
                    symbolLinks.type = baseType;
                }
            }
        }
        // Check for deprecations here, first on symbol, then on type.  However,
        // don't raise deprecation when the usage site is also a deprecated
        // declaration.
        const declarationNode = getSymNode(sym);
        if (declarationNode && mapper === undefined && isType(baseType)) {
            if (!isTypeReferenceContextDeprecated(node.parent)) {
                checkDeprecated(baseType, declarationNode, node);
            }
        }
        // Elements that could be used as type or values depending on the context
        if ((isType(baseType) && (baseType.kind === "EnumMember" || baseType.kind === "UnionVariant")) ||
            isNullType(baseType)) {
            return createIndeterminateEntity(baseType);
        }
        return baseType;
    }
    /**
     * Get or check the declared type of a templatable node.
     * @param node Declaration node
     * @param sym Node Symbol
     * @param mapper Type mapper for template resolution
     * @returns The declared type for the given node.
     */
    function getOrCheckDeclaredType(sym, decl, mapper) {
        const symbolLinks = getSymbolLinks(sym);
        if (symbolLinks.declaredType) {
            return symbolLinks.declaredType;
        }
        if (sym.flags & 4194304 /* SymbolFlags.LateBound */) {
            compilerAssert(sym.type, "Expected late bound symbol to have type");
            return sym.type;
        }
        if (sym.flags & 65536 /* SymbolFlags.Member */) {
            return checkMemberSym(sym, mapper);
        }
        else {
            return checkDeclaredType(sym, decl, mapper);
        }
    }
    /**
     * Check the declared type of a templatable node.
     * @param node Declaration node
     * @param sym Node Symbol
     * @param mapper Type mapper for template resolution
     * @returns The declared type for the given node.
     */
    function checkDeclaredTypeOrIndeterminate(sym, node, mapper) {
        const type = sym.flags & 2 /* SymbolFlags.Model */
            ? checkModelStatement(node, mapper)
            : sym.flags & 4 /* SymbolFlags.Scalar */
                ? checkScalar(node, mapper)
                : sym.flags & 128 /* SymbolFlags.Alias */
                    ? checkAlias(node, mapper)
                    : sym.flags & 32 /* SymbolFlags.Interface */
                        ? checkInterface(node, mapper)
                        : sym.flags & 8 /* SymbolFlags.Operation */
                            ? checkOperation(node, mapper)
                            : checkUnion(node, mapper);
        return type;
    }
    function checkDeclaredType(sym, node, mapper) {
        return getTypeForTypeOrIndeterminate(checkDeclaredTypeOrIndeterminate(sym, node, mapper));
    }
    function getOrInstantiateTemplate(templateNode, params, args, source, parentMapper, instantiateTempalates = true) {
        const symbolLinks = templateNode.kind === SyntaxKind.OperationStatement &&
            templateNode.parent.kind === SyntaxKind.InterfaceStatement
            ? getSymbolLinksForMember(templateNode)
            : getSymbolLinks(templateNode.symbol);
        compilerAssert(symbolLinks, `Unexpected checker error. symbolLinks was not defined for ${SyntaxKind[templateNode.kind]}`);
        if (symbolLinks.instantiations === undefined) {
            const type = getTypeForNode(templateNode);
            if (isErrorType(type)) {
                return errorType;
            }
            else {
                compilerAssert(false, `Unexpected checker error. symbolLinks.instantiations was not defined for ${SyntaxKind[templateNode.kind]}`);
            }
        }
        const mapper = createTypeMapper(params, args, source, parentMapper);
        const cached = symbolLinks.instantiations?.get(mapper.args);
        if (cached) {
            return cached;
        }
        if (instantiateTempalates) {
            return instantiateTemplate(symbolLinks.instantiations, templateNode, params, mapper);
        }
        else {
            return errorType;
        }
    }
    /**
     * Builds a model type from a template and its template arguments.
     * Adds the template node to a set we can check when we bind template
     * parameters to access type type arguments.
     *
     * This will fall over if the same template is ever being instantiated
     * twice at the same time, or if template parameters from more than one template
     * are ever in scope at once.
     */
    function instantiateTemplate(instantiations, templateNode, params, mapper) {
        const type = getTypeForNode(templateNode, mapper);
        if (!instantiations.get(mapper.args)) {
            instantiations.set(mapper.args, type);
        }
        if (type.kind === "Model") {
            type.templateNode = templateNode;
        }
        return type;
    }
    /** Check a union expresion used in a parameter constraint, those allow the use of `valueof` as a variant. */
    function checkMixedParameterConstraintUnion(node, mapper) {
        const values = [];
        const types = [];
        for (const option of node.options) {
            const [kind, type] = getTypeOrValueOfTypeForNode(option, mapper);
            if (kind === "value") {
                values.push(type);
            }
            else {
                types.push(type);
            }
        }
        return {
            entityKind: "MixedParameterConstraint",
            node,
            valueType: values.length === 0
                ? undefined
                : values.length === 1
                    ? values[0]
                    : createConstraintUnion(node, values),
            type: types.length === 0
                ? undefined
                : types.length === 1
                    ? types[0]
                    : createConstraintUnion(node, types),
        };
    }
    function createConstraintUnion(node, options) {
        const variants = createRekeyableMap();
        const union = createAndFinishType({
            kind: "Union",
            node,
            options,
            decorators: [],
            variants,
            expression: true,
        });
        for (const option of options) {
            const name = Symbol("indexer-union-variant");
            variants.set(name, createAndFinishType({
                kind: "UnionVariant",
                node: undefined,
                type: option,
                name,
                union,
                decorators: [],
            }));
        }
        return union;
    }
    function checkUnionExpression(node, mapper) {
        const unionType = createAndFinishType({
            kind: "Union",
            node,
            get options() {
                return Array.from(this.variants.values()).map((v) => v.type);
            },
            expression: true,
            variants: createRekeyableMap(),
            decorators: [],
        });
        for (const o of node.options) {
            const type = getTypeForNode(o, mapper);
            // The type `A | never` is just `A`
            if (type === neverType) {
                continue;
            }
            if (type.kind === "Union" && type.expression) {
                for (const [name, variant] of type.variants) {
                    unionType.variants.set(name, variant);
                }
            }
            else {
                const variant = createType({
                    kind: "UnionVariant",
                    type,
                    name: Symbol("name"),
                    decorators: [],
                    node: undefined,
                    union: unionType,
                });
                unionType.variants.set(variant.name, variant);
            }
        }
        linkMapper(unionType, mapper);
        return unionType;
    }
    /**
     * Intersection produces a model type from the properties of its operands.
     * So this doesn't work if we don't have a known set of properties (e.g.
     * with unions). The resulting model is anonymous.
     */
    function checkIntersectionExpression(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this model and we've already checked it
            return links.declaredType;
        }
        const options = node.options.map((o) => [o, getTypeForNode(o, mapper)]);
        const type = mergeModelTypes(node.symbol, node, options, mapper);
        linkType(links, type, mapper);
        return type;
    }
    function checkDecoratorDeclaration(node, mapper) {
        const symbol = getMergedSymbol(node.symbol);
        const links = getSymbolLinks(symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this operation and we've already checked it
            return links.declaredType;
        }
        const namespace = getParentNamespaceType(node);
        compilerAssert(namespace, `Decorator ${node.id.sv} should have resolved a namespace or found the global namespace.`);
        const name = node.id.sv;
        if (!(node.modifierFlags & 2 /* ModifierFlags.Extern */)) {
            reportCheckerDiagnostic(createDiagnostic({ code: "decorator-extern", target: node }));
        }
        const implementation = symbol.value;
        if (implementation === undefined) {
            reportCheckerDiagnostic(createDiagnostic({ code: "missing-implementation", target: node }));
        }
        const decoratorType = createType({
            kind: "Decorator",
            name: `@${name}`,
            namespace,
            node,
            target: checkFunctionParameter(node.target, mapper, true),
            parameters: node.parameters.map((x) => checkFunctionParameter(x, mapper, true)),
            implementation: implementation ?? (() => { }),
        });
        namespace.decoratorDeclarations.set(name, decoratorType);
        linkType(links, decoratorType, mapper);
        return decoratorType;
    }
    function checkFunctionDeclaration(node, mapper) {
        reportCheckerDiagnostic(createDiagnostic({ code: "function-unsupported", target: node }));
        return errorType;
    }
    function checkFunctionParameter(node, mapper, mixed) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType) {
            return links.declaredType;
        }
        if (node.rest &&
            node.type &&
            !(node.type.kind === SyntaxKind.ArrayExpression ||
                (node.type.kind === SyntaxKind.ValueOfExpression &&
                    node.type.target.kind === SyntaxKind.ArrayExpression))) {
            reportCheckerDiagnostic(createDiagnostic({ code: "rest-parameter-array", target: node.type }));
        }
        const base = {
            kind: "FunctionParameter",
            node,
            name: node.id.sv,
            optional: node.optional,
            rest: node.rest,
            implementation: node.symbol.value,
        };
        let parameterType;
        if (mixed) {
            const type = node.type
                ? getParamConstraintEntityForNode(node.type)
                : {
                    entityKind: "MixedParameterConstraint",
                    type: unknownType,
                };
            parameterType = createType({
                ...base,
                type,
                mixed: true,
                implementation: node.symbol.value,
            });
        }
        else {
            parameterType = createType({
                ...base,
                mixed: false,
                type: node.type ? getTypeForNode(node.type) : unknownType,
                implementation: node.symbol.value,
            });
        }
        linkType(links, parameterType, mapper);
        return parameterType;
    }
    function getTypeOrValueOfTypeForNode(node, mapper) {
        switch (node.kind) {
            case SyntaxKind.ValueOfExpression:
                const target = getTypeForNode(node.target, mapper);
                return ["value", target];
            default:
                return ["type", getTypeForNode(node, mapper)];
        }
    }
    function getParamConstraintEntityForNode(node, mapper) {
        switch (node.kind) {
            case SyntaxKind.UnionExpression:
                return checkMixedParameterConstraintUnion(node, mapper);
            default:
                const [kind, entity] = getTypeOrValueOfTypeForNode(node, mapper);
                return {
                    entityKind: "MixedParameterConstraint",
                    node: node,
                    type: kind === "value" ? undefined : entity,
                    valueType: kind === "value" ? entity : undefined,
                };
        }
    }
    function mergeModelTypes(parentModelSym, node, options, mapper) {
        const properties = createRekeyableMap();
        const intersection = createType({
            kind: "Model",
            node,
            name: "",
            namespace: getParentNamespaceType(node),
            properties: properties,
            decorators: [],
            derivedModels: [],
            sourceModels: [],
        });
        const indexers = [];
        const modelOptions = options.filter((entry) => {
            const [optionNode, option] = entry;
            if (option.kind === "TemplateParameter") {
                return false;
            }
            if (option.kind !== "Model") {
                reportCheckerDiagnostic(createDiagnostic({ code: "intersect-non-model", target: optionNode }));
                return false;
            }
            return true;
        });
        for (const [optionNode, option] of modelOptions) {
            if (option.indexer) {
                if (option.indexer.key.name === "integer") {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "intersect-invalid-index",
                        messageId: "array",
                        target: optionNode,
                    }));
                }
                else {
                    indexers.push(option.indexer);
                }
            }
        }
        for (const [_, option] of modelOptions) {
            intersection.sourceModels.push({ usage: "intersection", model: option });
            const allProps = walkPropertiesInherited(option);
            for (const prop of allProps) {
                if (properties.has(prop.name)) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "intersect-duplicate-property",
                        format: { propName: prop.name },
                        target: node,
                    }));
                    continue;
                }
                const memberSym = parentModelSym && getMemberSymbol(parentModelSym, prop.name);
                const overrides = {
                    sourceProperty: prop,
                    model: intersection,
                };
                const newPropType = memberSym
                    ? cloneTypeForSymbol(memberSym, prop, overrides)
                    : cloneType(prop, overrides);
                properties.set(prop.name, newPropType);
                linkIndirectMember(node, newPropType, mapper);
                for (const indexer of indexers.filter((x) => x !== option.indexer)) {
                    checkPropertyCompatibleWithIndexer(indexer, prop, node);
                }
            }
        }
        if (indexers.length === 1) {
            intersection.indexer = indexers[0];
        }
        else if (indexers.length > 1) {
            intersection.indexer = {
                key: indexers[0].key,
                value: mergeModelTypes(undefined, node, indexers.map((x) => [x.value.node, x.value]), mapper),
            };
        }
        linkMapper(intersection, mapper);
        return finishType(intersection);
    }
    function checkArrayExpression(node, mapper) {
        const elementType = getTypeForNode(node.elementType, mapper);
        const arrayType = getStdType("Array");
        const arrayNode = arrayType.node;
        const param = getTypeForNode(arrayNode.templateParameters[0]);
        return getOrInstantiateTemplate(arrayNode, [param], [elementType], { node, mapper }, undefined);
    }
    function checkNamespace(node) {
        const links = getSymbolLinks(getMergedSymbol(node.symbol));
        let type = links.type;
        if (!type) {
            type = initializeTypeForNamespace(node);
        }
        if (node.kind === SyntaxKind.NamespaceStatement) {
            if (isArray(node.statements)) {
                node.statements.forEach((x) => checkNode(x));
            }
            else if (node.statements) {
                const subNs = checkNamespace(node.statements);
                type.namespaces.set(subNs.name, subNs);
            }
        }
        return type;
    }
    function initializeTypeForNamespace(node) {
        compilerAssert(node.symbol, "Namespace is unbound.", node);
        const mergedSymbol = getMergedSymbol(node.symbol);
        const symbolLinks = getSymbolLinks(mergedSymbol);
        if (!symbolLinks.type) {
            // haven't seen this namespace before
            const namespace = getParentNamespaceType(node);
            const name = node.id.sv;
            const type = createType({
                kind: "Namespace",
                name,
                namespace,
                node: node,
                models: new Map(),
                scalars: new Map(),
                operations: new Map(),
                namespaces: new Map(),
                interfaces: new Map(),
                unions: new Map(),
                enums: new Map(),
                decoratorDeclarations: new Map(),
                functionDeclarations: new Map(),
                decorators: [],
            });
            symbolLinks.type = type;
            for (const sourceNode of mergedSymbol.declarations) {
                // namespaces created from TypeSpec scripts don't have decorators
                if (sourceNode.kind !== SyntaxKind.NamespaceStatement)
                    continue;
                type.decorators = type.decorators.concat(checkDecorators(type, sourceNode, undefined));
            }
            finishType(type);
            namespace?.namespaces.set(name, type);
        }
        return symbolLinks.type;
    }
    function getParentNamespaceType(node) {
        if (node === globalNamespaceType.node)
            return undefined;
        if (node.kind === SyntaxKind.ModelExpression ||
            node.kind === SyntaxKind.IntersectionExpression) {
            let parent = node.parent;
            while (parent !== undefined) {
                if (parent.kind === SyntaxKind.AliasStatement ||
                    parent.kind === SyntaxKind.ModelStatement ||
                    parent.kind === SyntaxKind.ScalarStatement ||
                    parent.kind === SyntaxKind.OperationStatement ||
                    parent.kind === SyntaxKind.EnumStatement ||
                    parent.kind === SyntaxKind.InterfaceStatement ||
                    parent.kind === SyntaxKind.UnionStatement ||
                    parent.kind === SyntaxKind.ModelExpression ||
                    parent.kind === SyntaxKind.IntersectionExpression) {
                    return getParentNamespaceType(parent);
                }
                else {
                    parent = parent.parent;
                }
            }
            return undefined;
        }
        if (node.kind === SyntaxKind.OperationStatement &&
            node.parent &&
            node.parent.kind === SyntaxKind.InterfaceStatement) {
            return getParentNamespaceType(node.parent);
        }
        if (!node.symbol.parent) {
            return globalNamespaceType;
        }
        if (node.symbol.parent.declarations[0].kind === SyntaxKind.TypeSpecScript ||
            node.symbol.parent.declarations[0].kind === SyntaxKind.JsSourceFile) {
            return globalNamespaceType;
        }
        const mergedSymbol = getMergedSymbol(node.symbol.parent);
        const symbolLinks = getSymbolLinks(mergedSymbol);
        if (!symbolLinks.type) {
            // in general namespaces should be typed before anything calls this function.
            // However, one case where this is not true is when a decorator on a namespace
            // refers to a model in another namespace. In this case, we need to evaluate
            // the namespace here.
            const namespaceNode = mergedSymbol.declarations.find((x) => x.kind === SyntaxKind.NamespaceStatement || x.kind === SyntaxKind.JsNamespaceDeclaration);
            compilerAssert(namespaceNode, "Can't find namespace declaration node.", node);
            symbolLinks.type = initializeTypeForNamespace(namespaceNode);
        }
        return symbolLinks.type;
    }
    function checkOperation(node, mapper, parentInterface) {
        const inInterface = node.parent?.kind === SyntaxKind.InterfaceStatement;
        const symbol = inInterface ? getSymbolForMember(node) : node.symbol;
        const links = symbol && getSymbolLinks(symbol);
        if (links) {
            if (links.declaredType && mapper === undefined) {
                // we're not instantiating this operation and we've already checked it
                return links.declaredType;
            }
        }
        if (mapper === undefined && inInterface) {
            compilerAssert(parentInterface, "Operation in interface should already have been checked.", node.parent);
        }
        checkTemplateDeclaration(node, mapper);
        // If we are instantating operation inside of interface
        if (isTemplatedNode(node) && mapper !== undefined && parentInterface) {
            mapper = { ...mapper, partial: true };
        }
        const namespace = getParentNamespaceType(node);
        const name = node.id.sv;
        let decorators = [];
        const { resolvedSymbol: parameterModelSym } = resolver.resolveMetaMemberByName(symbol, "parameters");
        if (parameterModelSym?.members) {
            const members = resolver.getAugmentedSymbolTable(parameterModelSym.members);
            const paramDocs = extractParamDocs(node);
            for (const [name, memberSym] of members) {
                const doc = paramDocs.get(name);
                if (doc) {
                    docFromCommentForSym.set(memberSym, doc);
                }
            }
        }
        // Is this a definition or reference?
        let parameters, returnType, sourceOperation;
        if (node.signature.kind === SyntaxKind.OperationSignatureReference) {
            // Attempt to resolve the operation
            const baseOperation = checkOperationIs(node, node.signature.baseOperation, mapper);
            if (baseOperation) {
                sourceOperation = baseOperation;
                // Reference the same return type and create the parameters type
                const clone = initializeClone(baseOperation.parameters, {
                    properties: createRekeyableMap(),
                });
                clone.properties = createRekeyableMap(Array.from(baseOperation.parameters.properties.entries()).map(([key, prop]) => [
                    key,
                    cloneTypeForSymbol(getMemberSymbol(parameterModelSym, prop.name), prop, {
                        model: clone,
                        sourceProperty: prop,
                    }),
                ]));
                parameters = finishType(clone);
                returnType = baseOperation.returnType;
                // Copy decorators from the base operation, inserting the base decorators first
                decorators = [...baseOperation.decorators];
            }
            else {
                // If we can't resolve the signature we return an empty model.
                parameters = createAndFinishType({
                    kind: "Model",
                    name: "",
                    decorators: [],
                    properties: createRekeyableMap(),
                    derivedModels: [],
                    sourceModels: [],
                });
                returnType = voidType;
            }
        }
        else {
            parameters = getTypeForNode(node.signature.parameters, mapper);
            returnType = getTypeForNode(node.signature.returnType, mapper);
        }
        const operationType = createType({
            kind: "Operation",
            name,
            namespace,
            node,
            parameters,
            returnType,
            decorators,
            sourceOperation,
            interface: parentInterface,
        });
        if (links) {
            linkType(links, operationType, mapper);
        }
        decorators.push(...checkDecorators(operationType, node, mapper));
        operationType.parameters.namespace = namespace;
        const parent = node.parent;
        linkMapper(operationType, mapper);
        if (parent.kind === SyntaxKind.InterfaceStatement) {
            if (shouldCreateTypeForTemplate(parent, mapper) &&
                shouldCreateTypeForTemplate(node, mapper)) {
                finishType(operationType);
            }
        }
        else {
            if (shouldCreateTypeForTemplate(node, mapper)) {
                finishType(operationType);
            }
            if (mapper === undefined) {
                namespace?.operations.set(name, operationType);
            }
        }
        return operationType;
    }
    function checkOperationIs(operation, opReference, mapper) {
        if (!opReference)
            return undefined;
        // Ensure that we don't end up with a circular reference to the same operation
        const opSymId = getNodeSym(operation);
        if (opSymId) {
            pendingResolutions.start(opSymId, ResolutionKind.BaseType);
        }
        const target = resolver.getNodeLinks(opReference).resolvedSymbol;
        // Did we encounter a circular operation reference?
        if (target && pendingResolutions.has(target, ResolutionKind.BaseType)) {
            if (mapper === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "circular-op-signature",
                    format: { typeName: target.name },
                    target: opReference,
                }));
            }
            return undefined;
        }
        // Resolve the base operation type
        const baseOperation = getTypeForNode(opReference, mapper);
        if (opSymId) {
            pendingResolutions.finish(opSymId, ResolutionKind.BaseType);
        }
        if (isErrorType(baseOperation)) {
            return undefined;
        }
        // Was the wrong type referenced?
        if (baseOperation.kind !== "Operation") {
            reportCheckerDiagnostic(createDiagnostic({ code: "is-operation", target: opReference }));
            return;
        }
        return baseOperation;
    }
    function getGlobalNamespaceType() {
        return globalNamespaceType;
    }
    function getGlobalNamespaceNode() {
        return resolver.symbols.global.declarations[0];
    }
    function checkTupleExpression(node, mapper) {
        return createAndFinishType({
            kind: "Tuple",
            node: node,
            values: node.values.map((v) => getTypeForNode(v, mapper)),
        });
    }
    function getSymbolLinks(s) {
        return resolver.getSymbolLinks(s);
    }
    function resolveRelatedSymbols(id, mapper) {
        let sym;
        const { node, kind } = getIdentifierContext(id);
        switch (kind) {
            case IdentifierKind.ModelExpressionProperty:
            case IdentifierKind.ObjectLiteralProperty:
                const model = getReferencedModel(node);
                return model
                    .map((m) => getMemberSymbol(m.node.symbol, id.sv))
                    .filter((m) => m !== undefined);
            case IdentifierKind.ModelStatementProperty:
            case IdentifierKind.Declaration:
                const links = resolver.getNodeLinks(id);
                return links.resolvedSymbol === undefined ? undefined : [links.resolvedSymbol];
            case IdentifierKind.Other:
                return undefined;
            case IdentifierKind.Decorator:
            case IdentifierKind.Function:
            case IdentifierKind.Using:
            case IdentifierKind.TypeReference:
                let ref = id;
                let resolveDecorator = kind === IdentifierKind.Decorator;
                if (id.parent?.kind === SyntaxKind.MemberExpression) {
                    if (id.parent.id === id) {
                        // If the identifier is Y in X.Y, then resolve (X.Y).
                        ref = id.parent;
                    }
                    else {
                        // If the identifier is X in X.Y then we are resolving a
                        // namespace, which is never a decorator.
                        resolveDecorator = false;
                    }
                }
                sym = resolveTypeReferenceSym(ref, mapper, resolveDecorator);
                break;
            case IdentifierKind.TemplateArgument:
                const templates = getTemplateDeclarationsForArgument(node, mapper);
                const firstMatchingParameter = templates
                    .flatMap((t) => t.templateParameters)
                    .find((p) => p.id.sv === id.sv);
                if (firstMatchingParameter) {
                    sym = getMergedSymbol(firstMatchingParameter.symbol);
                }
                break;
            default:
                const _assertNever = kind;
                compilerAssert(false, "Unreachable");
        }
        if (sym) {
            if (sym.symbolSource) {
                return [sym.symbolSource];
            }
            else {
                return [sym];
            }
        }
        return undefined; //sym?.symbolSource ?? sym;
    }
    function getTemplateDeclarationsForArgument(node, mapper) {
        const ref = node.parent;
        let resolved = resolveTypeReferenceSym(ref, mapper, false);
        // if the reference type can't be resolved and has parse error,
        // it likely means the reference type hasn't been completed yet. i.e. Foo<string,
        // so try to resolve it by it's target directly to see if we can find its sym
        if (!resolved && hasParseError(ref) && ref.target !== undefined) {
            resolved = resolveTypeReferenceSym(ref.target, mapper, false);
        }
        return (resolved?.declarations.filter((n) => isTemplatedNode(n)) ?? []);
    }
    function getReferencedModel(propertyNode) {
        const isModelOrArrayValue = (n) => n?.kind === SyntaxKind.ArrayLiteral || n?.kind === SyntaxKind.ObjectLiteral;
        const isModelOrArrayType = (n) => n?.kind === SyntaxKind.ModelExpression || n?.kind === SyntaxKind.TupleExpression;
        const isModelOrArray = (n) => isModelOrArrayValue(n) || isModelOrArrayType(n);
        const path = [];
        let preNode;
        const foundNode = getFirstAncestor(propertyNode, (n) => {
            pushToModelPath(n, preNode, path);
            preNode = n;
            return ((isModelOrArray(n) &&
                (n.parent?.kind === SyntaxKind.TemplateParameterDeclaration ||
                    n.parent?.kind === SyntaxKind.TemplateArgument ||
                    n.parent?.kind === SyntaxKind.DecoratorExpression)) ||
                (isModelOrArrayValue(n) &&
                    (n.parent?.kind === SyntaxKind.CallExpression ||
                        n.parent?.kind === SyntaxKind.ConstStatement)));
        });
        let refType;
        switch (foundNode?.parent?.kind) {
            case SyntaxKind.TemplateParameterDeclaration:
            case SyntaxKind.TemplateArgument:
                refType = getReferencedTypeFromTemplateDeclaration(foundNode);
                break;
            case SyntaxKind.DecoratorExpression:
                refType = getReferencedTypeFromDecoratorArgument(foundNode);
                break;
            case SyntaxKind.CallExpression:
                refType = getReferencedTypeFromScalarConstructor(foundNode);
                break;
            case SyntaxKind.ConstStatement:
                refType = getReferencedTypeFromConstAssignment(foundNode);
                break;
        }
        return getNestedModel(refType, path);
        function pushToModelPath(node, preNode, path) {
            if (node.kind === SyntaxKind.ArrayLiteral || node.kind === SyntaxKind.TupleExpression) {
                const index = node.values.findIndex((n) => n === preNode);
                if (index >= 0) {
                    path.unshift({ tupleIndex: index });
                }
                else {
                    compilerAssert(false, "not expected, can't find child from the parent?");
                }
            }
            if (node.kind === SyntaxKind.ModelProperty ||
                node.kind === SyntaxKind.ObjectLiteralProperty) {
                path.unshift({ propertyName: node.id.sv });
            }
        }
        function getNestedModel(modelOrTupleOrUnion, path) {
            let cur = modelOrTupleOrUnion;
            if (cur && cur.kind !== "Model" && cur.kind !== "Tuple" && cur.kind !== "Union") {
                return [];
            }
            if (path.length === 0) {
                // Handle union and model type nesting when path is empty
                switch (cur?.kind) {
                    case "Model":
                        return [cur];
                    case "Union":
                        const models = [];
                        for (const variant of cur.variants.values()) {
                            if (variant.type.kind === "Model" ||
                                variant.type.kind === "Tuple" ||
                                variant.type.kind === "Union") {
                                models.push(...(getNestedModel(variant.type, path) ?? []));
                            }
                        }
                        return models;
                    default:
                        return [];
                }
            }
            const seg = path[0];
            switch (cur?.kind) {
                case "Tuple":
                    if (seg.tupleIndex !== undefined &&
                        seg.tupleIndex >= 0 &&
                        seg.tupleIndex < cur.values.length) {
                        return getNestedModel(cur.values[seg.tupleIndex], path.slice(1));
                    }
                    else {
                        return [];
                    }
                case "Model":
                    if (cur.name === "Array" && seg.tupleIndex !== undefined) {
                        cur = cur.templateMapper?.args[0];
                    }
                    else if (cur.name !== "Array" && seg.propertyName) {
                        cur = cur.properties.get(seg.propertyName)?.type;
                    }
                    else {
                        return [];
                    }
                    return getNestedModel(cur, path.slice(1));
                case "Union":
                    // When seg.property name exists, it means that it is in the union model or tuple,
                    // and the corresponding model or tuple needs to be found recursively.
                    const models = [];
                    for (const variant of cur.variants.values()) {
                        if (variant.type.kind === "Model" ||
                            variant.type.kind === "Tuple" ||
                            variant.type.kind === "Union") {
                            models.push(...(getNestedModel(variant.type, path) ?? []));
                        }
                    }
                    return models;
                default:
                    return [];
            }
        }
        function getReferencedTypeFromTemplateDeclaration(node) {
            let templateParmaeterDeclNode = undefined;
            if (node?.parent?.kind === SyntaxKind.TemplateArgument &&
                node?.parent?.parent?.kind === SyntaxKind.TypeReference) {
                const argNode = node.parent;
                const refNode = node.parent.parent;
                const decl = getTemplateDeclarationsForArgument(argNode, 
                // We should be giving the argument so the mapper here should be undefined
                undefined /* mapper */);
                const index = refNode.arguments.findIndex((n) => n === argNode);
                if (decl.length > 0 && decl[0].templateParameters.length > index) {
                    templateParmaeterDeclNode = decl[0].templateParameters[index];
                }
            }
            else if (node.parent?.kind === SyntaxKind.TemplateParameterDeclaration) {
                templateParmaeterDeclNode = node?.parent;
            }
            if (templateParmaeterDeclNode?.kind !== SyntaxKind.TemplateParameterDeclaration ||
                !templateParmaeterDeclNode.constraint) {
                return undefined;
            }
            let constraintType;
            if (isModelOrArrayValue(node) &&
                templateParmaeterDeclNode.constraint.kind === SyntaxKind.ValueOfExpression) {
                constraintType = program.checker.getTypeForNode(templateParmaeterDeclNode.constraint.target);
            }
            else if (isModelOrArrayType(node) &&
                templateParmaeterDeclNode.constraint.kind !== SyntaxKind.ValueOfExpression) {
                constraintType = program.checker.getTypeForNode(templateParmaeterDeclNode.constraint);
            }
            return constraintType;
        }
        function getReferencedTypeFromScalarConstructor(argNode) {
            const callExpNode = argNode?.parent;
            if (callExpNode?.kind !== SyntaxKind.CallExpression) {
                return undefined;
            }
            const ctorType = checkCallExpressionTarget(callExpNode, undefined);
            if (ctorType?.kind !== "ScalarConstructor") {
                return undefined;
            }
            const argIndex = callExpNode.arguments.findIndex((n) => n === argNode);
            if (argIndex < 0 || argIndex >= ctorType.parameters.length) {
                return undefined;
            }
            const arg = ctorType.parameters[argIndex];
            return arg.type;
        }
        function getReferencedTypeFromConstAssignment(valueNode) {
            const constNode = valueNode?.parent;
            if (!constNode ||
                constNode.kind !== SyntaxKind.ConstStatement ||
                !constNode.type ||
                constNode.value !== valueNode) {
                return undefined;
            }
            return program.checker.getTypeForNode(constNode.type);
        }
        function getReferencedTypeFromDecoratorArgument(decArgNode) {
            const decNode = decArgNode?.parent;
            if (decNode?.kind !== SyntaxKind.DecoratorExpression) {
                return undefined;
            }
            const decSym = program.checker.resolveRelatedSymbols(decNode.target.kind === SyntaxKind.MemberExpression ? decNode.target.id : decNode.target);
            if (!decSym || decSym.length <= 0) {
                return undefined;
            }
            const decDecl = decSym[0].declarations.find((x) => x.kind === SyntaxKind.DecoratorDeclarationStatement);
            if (!decDecl) {
                return undefined;
            }
            const decType = program.checker.getTypeForNode(decDecl);
            compilerAssert(decType.kind === "Decorator", "Expected type to be a Decorator.");
            const argIndex = decNode.arguments.findIndex((n) => n === decArgNode);
            if (argIndex < 0 || argIndex >= decType.parameters.length) {
                return undefined;
            }
            const decArg = decType.parameters[argIndex];
            let type;
            if (isModelOrArrayValue(decArgNode)) {
                type = decArg.type.valueType;
            }
            else if (isModelOrArrayType(decArgNode)) {
                type = decArg.type.type ?? decArg.type.valueType;
            }
            else {
                compilerAssert(false, "not expected node type to get reference model from decorator argument");
            }
            return type;
        }
    }
    function resolveCompletions(identifier) {
        const completions = new Map();
        const { kind, node: ancestor } = getIdentifierContext(identifier);
        switch (kind) {
            case IdentifierKind.Using:
            case IdentifierKind.Decorator:
            case IdentifierKind.Function:
            case IdentifierKind.TypeReference:
            case IdentifierKind.ModelExpressionProperty:
            case IdentifierKind.ModelStatementProperty:
            case IdentifierKind.ObjectLiteralProperty:
                break; // supported
            case IdentifierKind.Other:
                return completions; // not implemented
            case IdentifierKind.Declaration:
                return completions; // cannot complete, name can be chosen arbitrarily
            case IdentifierKind.TemplateArgument: {
                const templates = getTemplateDeclarationsForArgument(ancestor, undefined);
                for (const template of templates) {
                    for (const param of template.templateParameters) {
                        addCompletion(param.id.sv, param.symbol);
                    }
                }
                return completions;
            }
            default:
                const _assertNever = kind;
                compilerAssert(false, "Unreachable");
        }
        if (kind === IdentifierKind.ModelStatementProperty) {
            const model = ancestor.parent;
            const modelType = program.checker.getTypeForNode(model);
            const baseType = modelType.baseModel;
            const baseNode = baseType?.node;
            if (!baseNode) {
                return completions;
            }
            for (const prop of baseType.properties.values()) {
                if (identifier.sv === prop.name || !modelType.properties.has(prop.name)) {
                    const sym = getMemberSymbol(baseNode.symbol, prop.name);
                    if (sym) {
                        addCompletion(prop.name, sym);
                    }
                }
            }
        }
        else if (kind === IdentifierKind.ModelExpressionProperty ||
            kind === IdentifierKind.ObjectLiteralProperty) {
            const model = getReferencedModel(ancestor);
            if (model.length <= 0) {
                return completions;
            }
            const curModelNode = ancestor.parent;
            for (const curModel of model) {
                for (const prop of walkPropertiesInherited(curModel)) {
                    if (identifier.sv === prop.name ||
                        !curModelNode.properties.find((p) => (p.kind === SyntaxKind.ModelProperty ||
                            p.kind === SyntaxKind.ObjectLiteralProperty) &&
                            p.id.sv === prop.name)) {
                        const sym = getMemberSymbol(curModel.node.symbol, prop.name);
                        if (sym) {
                            addCompletion(prop.name, sym);
                        }
                    }
                }
            }
        }
        else if (identifier.parent && identifier.parent.kind === SyntaxKind.MemberExpression) {
            let base = resolver.getNodeLinks(identifier.parent.base).resolvedSymbol;
            if (base) {
                if (base.flags & 128 /* SymbolFlags.Alias */) {
                    base = getAliasedSymbol(base, undefined);
                }
                if (base) {
                    if (identifier.parent.selector === "::") {
                        if (base?.node === undefined && base?.declarations && base.declarations.length > 0) {
                            // Process meta properties separately, such as `::parameters`, `::returnType`
                            const nodeModels = base?.declarations[0];
                            if (nodeModels.kind === SyntaxKind.OperationStatement) {
                                const operation = nodeModels;
                                addCompletion("parameters", operation.symbol);
                                addCompletion("returnType", operation.symbol);
                            }
                        }
                        else if (base?.node?.kind === SyntaxKind.ModelProperty) {
                            // Process meta properties separately, such as `::type`
                            const metaProperty = base.node;
                            addCompletion("type", metaProperty.symbol);
                        }
                    }
                    else {
                        addCompletions(base.exports ?? base.members);
                    }
                }
            }
        }
        else {
            // We will only add template arguments if the template isn't already named
            // to avoid completing the name of the argument again.
            if (kind === IdentifierKind.TypeReference &&
                exprIsBareIdentifier(ancestor) &&
                ancestor.parent?.kind === SyntaxKind.TemplateArgument &&
                ancestor.parent.name === undefined) {
                const templates = getTemplateDeclarationsForArgument(ancestor.parent, undefined);
                for (const template of templates) {
                    for (const param of template.templateParameters) {
                        addCompletion(param.id.sv, param.symbol, { suffix: " = " });
                    }
                }
            }
            let scope = identifier.parent;
            while (scope && scope.kind !== SyntaxKind.TypeSpecScript) {
                if (scope.symbol && scope.symbol.exports) {
                    const mergedSymbol = getMergedSymbol(scope.symbol);
                    addCompletions(mergedSymbol.exports);
                }
                if ("locals" in scope) {
                    addCompletions(scope.locals);
                }
                scope = scope.parent;
            }
            if (scope && scope.kind === SyntaxKind.TypeSpecScript) {
                // check any blockless namespace decls
                for (const ns of scope.inScopeNamespaces) {
                    const mergedSymbol = getMergedSymbol(ns.symbol);
                    addCompletions(mergedSymbol.exports);
                }
                // check "global scope" declarations
                addCompletions(resolver.symbols.global.exports);
                // check "global scope" usings
                addCompletions(scope.locals);
            }
        }
        return completions;
        function addCompletions(table) {
            if (!table) {
                return;
            }
            table = resolver.getAugmentedSymbolTable(table);
            for (const [key, sym] of table) {
                if (sym.flags & 16384 /* SymbolFlags.DuplicateUsing */) {
                    const duplicates = table.duplicates.get(sym);
                    for (const duplicate of duplicates) {
                        if (duplicate.flags & 8192 /* SymbolFlags.Using */) {
                            const fqn = getFullyQualifiedSymbolName(duplicate.symbolSource);
                            addCompletion(fqn, duplicate);
                        }
                    }
                }
                else {
                    addCompletion(key, sym);
                }
            }
        }
        function addCompletion(key, sym, options = {}) {
            if (sym.symbolSource) {
                sym = sym.symbolSource;
            }
            if (!shouldAddCompletion(sym)) {
                return;
            }
            if (key.startsWith("@")) {
                key = key.slice(1);
            }
            if (!completions.has(key)) {
                completions.set(key, { ...options, sym });
            }
        }
        function shouldAddCompletion(sym) {
            switch (kind) {
                case IdentifierKind.ModelExpressionProperty:
                case IdentifierKind.ModelStatementProperty:
                case IdentifierKind.ObjectLiteralProperty:
                    return !!(sym.flags & 65536 /* SymbolFlags.Member */);
                case IdentifierKind.Decorator:
                    // Only return decorators and namespaces when completing decorator
                    return !!(sym.flags & (512 /* SymbolFlags.Decorator */ | 256 /* SymbolFlags.Namespace */));
                case IdentifierKind.Using:
                    // Only return namespaces when completing using
                    return !!(sym.flags & 256 /* SymbolFlags.Namespace */);
                case IdentifierKind.TypeReference:
                    // Do not return functions or decorators when completing types
                    return !(sym.flags & (2048 /* SymbolFlags.Function */ | 512 /* SymbolFlags.Decorator */));
                case IdentifierKind.TemplateArgument:
                    return !!(sym.flags & 1024 /* SymbolFlags.TemplateParameter */);
                default:
                    compilerAssert(false, "We should have bailed up-front on other kinds.");
            }
        }
    }
    function getCodefixesForUnknownIdentifier(node) {
        switch (node.sv) {
            case "number":
                return [createChangeIdentifierCodeFix(node, "float64")];
            default:
                return undefined;
        }
    }
    function resolveTypeReferenceSym(node, mapper, options) {
        const resolvedOptions = typeof options === "boolean"
            ? { ...defaultSymbolResolutionOptions, resolveDecorators: options }
            : { ...defaultSymbolResolutionOptions, ...(options ?? {}) };
        if (mapper === undefined && resolvedOptions.checkTemplateTypes && referenceSymCache.has(node)) {
            return referenceSymCache.get(node);
        }
        const sym = resolveTypeReferenceSymInternal(node, mapper, resolvedOptions);
        if (resolvedOptions.checkTemplateTypes) {
            referenceSymCache.set(node, sym);
        }
        return sym;
    }
    function resolveTypeReferenceSymInternal(node, mapper, options) {
        if (hasParseError(node)) {
            // Don't report synthetic identifiers used for parser error recovery.
            // The parse error is the root cause and will already have been logged.
            return undefined;
        }
        if (node.kind === SyntaxKind.TypeReference) {
            return resolveTypeReferenceSym(node.target, mapper, options);
        }
        else if (node.kind === SyntaxKind.Identifier) {
            const links = resolver.getNodeLinks(node);
            if (mapper === undefined && links.resolutionResult) {
                if (mapper === undefined && // do not report error when instantiating
                    links.resolutionResult & (ResolutionResultFlags.NotFound | ResolutionResultFlags.Unknown)) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-ref",
                        messageId: options.resolveDecorators ? "decorator" : "identifier",
                        format: { id: printTypeReferenceNode(node) },
                        target: node,
                        codefixes: getCodefixesForUnknownIdentifier(node),
                    }));
                }
                else if (links.resolutionResult & ResolutionResultFlags.Ambiguous) {
                    reportAmbiguousIdentifier(node, links.ambiguousSymbols);
                }
            }
            const sym = links.resolvedSymbol;
            return sym?.symbolSource ?? sym;
        }
        else if (node.kind === SyntaxKind.MemberExpression) {
            let base = resolveTypeReferenceSym(node.base, mapper, {
                ...options,
                resolveDecorators: false, // when resolving decorator the base cannot also be one
            });
            if (!base) {
                return undefined;
            }
            // when resolving a type reference based on an alias, unwrap the alias.
            if (base.flags & 128 /* SymbolFlags.Alias */) {
                const aliasedSym = getAliasedSymbol(base, mapper);
                if (!aliasedSym) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-ref",
                        messageId: "node",
                        format: {
                            id: node.id.sv,
                            nodeName: base.declarations[0]
                                ? SyntaxKind[base.declarations[0].kind]
                                : "Unknown node",
                        },
                        target: node,
                    }));
                    return undefined;
                }
                base = aliasedSym;
            }
            return resolveMemberInContainer(base, node, options);
        }
        compilerAssert(false, `Unknown type reference kind "${SyntaxKind[node.kind]}"`, node);
    }
    function reportAmbiguousIdentifier(node, symbols) {
        const duplicateNames = symbols.map((s) => getFullyQualifiedSymbolName(s, { useGlobalPrefixAtTopLevel: true }));
        program.reportDiagnostic(createDiagnostic({
            code: "ambiguous-symbol",
            format: { name: node.sv, duplicateNames: duplicateNames.join(", ") },
            target: node,
        }));
    }
    function resolveMemberInContainer(base, node, options) {
        const { finalSymbol: sym, resolvedSymbol: nextSym } = resolver.resolveMemberExpressionForSym(base, node, options);
        const symbol = nextSym ?? sym;
        if (symbol) {
            return symbol;
        }
        if (base.flags & 256 /* SymbolFlags.Namespace */) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-ref",
                messageId: "underNamespace",
                format: {
                    namespace: getFullyQualifiedSymbolName(base),
                    id: node.id.sv,
                },
                target: node,
            }));
        }
        else if (base.flags & 512 /* SymbolFlags.Decorator */) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-ref",
                messageId: "inDecorator",
                format: { id: node.id.sv },
                target: node,
            }));
        }
        else if (base.flags & 2048 /* SymbolFlags.Function */) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-ref",
                messageId: "node",
                format: { id: node.id.sv, nodeName: "function" },
                target: node,
            }));
        }
        else if (base.flags & 118 /* SymbolFlags.MemberContainer */) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-ref",
                messageId: node.selector === "." ? "member" : "metaProperty",
                format: { kind: getMemberKindName(getSymNode(base)), id: node.id.sv },
                target: node,
            }));
        }
        else {
            const symNode = getSymNode(base);
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-ref",
                messageId: "node",
                format: {
                    id: node.id.sv,
                    nodeName: symNode ? SyntaxKind[symNode.kind] : "Unknown node",
                },
                target: node,
            }));
        }
        return undefined;
    }
    function getMemberKindName(node) {
        switch (node.kind) {
            case SyntaxKind.ModelStatement:
            case SyntaxKind.ModelExpression:
                return "Model";
            case SyntaxKind.ModelProperty:
                return "ModelProperty";
            case SyntaxKind.EnumStatement:
                return "Enum";
            case SyntaxKind.InterfaceStatement:
                return "Interface";
            case SyntaxKind.UnionStatement:
                return "Union";
            default:
                return "Type";
        }
    }
    /**
     * Return the symbol that is aliased by this alias declaration. If no such symbol is aliased,
     * return the symbol for the alias instead. For member containers which need to be late bound
     * (i.e. they contain symbols we don't know until we've instantiated the type and the type is an
     * instantiation) we late bind the container which creates the symbol that will hold its members.
     */
    function getAliasedSymbol(aliasSymbol, mapper) {
        const node = getSymNode(aliasSymbol);
        const links = resolver.getSymbolLinks(aliasSymbol);
        if (!links.aliasResolutionIsTemplate) {
            return links.aliasedSymbol ?? resolver.getNodeLinks(node).resolvedSymbol;
        }
        // Otherwise for templates we need to get the type and retrieve the late bound symbol.
        const aliasType = getTypeForNode(node, mapper);
        if (isErrorType(aliasType)) {
            return undefined;
        }
        switch (aliasType.kind) {
            case "Model":
            case "Interface":
            case "Union":
                if (isTemplateInstance(aliasType)) {
                    // this is an alias for some instantiation, so late-bind the instantiation
                    lateBindMemberContainer(aliasType);
                    return aliasType.symbol;
                }
            // fallthrough
            default:
                // get the symbol from the node aliased type's node, or just return the base
                // if it doesn't have a symbol (which will likely result in an error later on)
                return getMergedSymbol(aliasType.node.symbol) ?? aliasSymbol;
        }
    }
    function checkStringTemplateExpresion(node, mapper) {
        let hasType = false;
        let hasValue = false;
        const spanTypeOrValues = node.spans.map((span) => [span, checkNode(span.expression, mapper)]);
        for (const [_, typeOrValue] of spanTypeOrValues) {
            if (typeOrValue !== null) {
                if (isValue(typeOrValue)) {
                    hasValue = true;
                }
                else if ("kind" in typeOrValue && typeOrValue.kind === "TemplateParameter") {
                    if (typeOrValue.constraint) {
                        if (typeOrValue.constraint.valueType) {
                            hasValue = true;
                        }
                        if (typeOrValue.constraint.type) {
                            hasType = true;
                        }
                    }
                    else {
                        hasType = true;
                    }
                }
                else {
                    hasType = true;
                }
            }
        }
        if (hasType && hasValue) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "mixed-string-template",
                target: node,
            }));
            return null;
        }
        if (hasValue) {
            let str = node.head.value;
            for (const [span, typeOrValue] of spanTypeOrValues) {
                if (typeOrValue !== null &&
                    (!("kind" in typeOrValue) || typeOrValue.kind !== "TemplateParameter")) {
                    compilerAssert(typeOrValue !== null && isValue(typeOrValue), "Expected value.");
                    str += stringifyValueForTemplate(typeOrValue);
                }
                str += span.literal.value;
            }
            return checkStringValue(createLiteralType(str), undefined, node);
        }
        else {
            let hasNonStringElement = false;
            let stringValue = node.head.value;
            const spans = [createTemplateSpanLiteral(node.head)];
            for (const [span, typeOrValue] of spanTypeOrValues) {
                compilerAssert(typeOrValue !== null && !isValue(typeOrValue), "Expected type.");
                const type = typeOrValue.entityKind === "Indeterminate" ? typeOrValue.type : typeOrValue;
                const spanValue = createTemplateSpanValue(span.expression, type);
                spans.push(spanValue);
                const spanValueAsString = stringifyTypeForTemplate(type);
                if (spanValueAsString !== undefined) {
                    stringValue += spanValueAsString;
                }
                else {
                    hasNonStringElement = true;
                }
                spans.push(createTemplateSpanLiteral(span.literal));
                stringValue += span.literal.value;
            }
            return createIndeterminateEntity(createType({
                kind: "StringTemplate",
                node,
                spans,
                stringValue: hasNonStringElement ? undefined : stringValue,
            }));
        }
    }
    function createIndeterminateEntity(type) {
        const existing = indeterminateEntities.get(type);
        if (existing) {
            return existing;
        }
        const entity = {
            entityKind: "Indeterminate",
            type,
        };
        indeterminateEntities.set(type, entity);
        return entity;
    }
    function stringifyTypeForTemplate(type) {
        switch (type.kind) {
            case "String":
            case "Number":
            case "Boolean":
                return String(type.value);
            case "StringTemplate":
                if (type.stringValue !== undefined) {
                    return type.stringValue;
                }
                return undefined;
            default:
                return undefined;
        }
    }
    function stringifyValueForTemplate(value) {
        switch (value.valueKind) {
            case "StringValue":
            case "NumericValue":
            case "BooleanValue":
                return value.value.toString();
            default:
                reportCheckerDiagnostic(createDiagnostic({
                    code: "non-literal-string-template",
                    target: value,
                }));
                return `[${value.valueKind}]`;
        }
    }
    function createTemplateSpanLiteral(node) {
        return createType({
            kind: "StringTemplateSpan",
            node: node,
            isInterpolated: false,
            type: getLiteralType(node),
        });
    }
    function createTemplateSpanValue(node, type) {
        return createType({
            kind: "StringTemplateSpan",
            node: node,
            isInterpolated: true,
            type: type,
        });
    }
    function checkStringLiteral(str) {
        return createIndeterminateEntity(getLiteralType(str));
    }
    function checkNumericLiteral(num) {
        return createIndeterminateEntity(getLiteralType(num));
    }
    function checkBooleanLiteral(bool) {
        return createIndeterminateEntity(getLiteralType(bool));
    }
    function checkProgram() {
        checkDuplicateSymbols();
        for (const file of program.sourceFiles.values()) {
            checkDuplicateUsings(file);
            for (const ns of file.namespaces) {
                initializeTypeForNamespace(ns);
            }
        }
        for (const file of program.sourceFiles.values()) {
            checkSourceFile(file);
        }
        internalDecoratorValidation();
    }
    function checkDuplicateSymbols() {
        program.reportDuplicateSymbols(resolver.symbols.global.exports);
        for (const file of program.sourceFiles.values()) {
            for (const ns of file.namespaces) {
                const exports = getMergedSymbol(ns.symbol).exports ?? ns.symbol.exports;
                program.reportDuplicateSymbols(exports);
            }
        }
    }
    /** Report error with duplicate using in the same scope. */
    function checkDuplicateUsings(file) {
        const duplicateTrackers = new Map();
        function getTracker(sym) {
            const existing = duplicateTrackers.get(sym);
            if (existing)
                return existing;
            const newTacker = new DuplicateTracker();
            duplicateTrackers.set(sym, newTacker);
            return newTacker;
        }
        for (const using of file.usings) {
            const ns = using.parent;
            const sym = getMergedSymbol(ns.symbol);
            const tracker = getTracker(sym);
            const targetSym = resolver.getNodeLinks(using.name).resolvedSymbol;
            if (!targetSym)
                continue;
            tracker.track(targetSym, using);
        }
        for (const tracker of duplicateTrackers.values()) {
            for (const [_, nodes] of tracker.entries()) {
                for (const node of nodes) {
                    program.reportDiagnostic(createDiagnostic({
                        code: "duplicate-using",
                        format: { usingName: typeReferenceToString(node.name) },
                        target: node,
                    }));
                }
            }
        }
    }
    /**
     * Post checking validation for internal decorators.
     */
    function internalDecoratorValidation() {
        validateInheritanceDiscriminatedUnions(program);
    }
    function checkSourceFile(file) {
        for (const statement of file.statements) {
            checkNode(statement, undefined);
        }
    }
    /**
     * Check that the given node template parameters are valid if applicable.
     * @param node Node with template parameters
     * @param mapper Type mapper, set if instantiating the template, undefined otherwise.
     */
    function checkTemplateDeclaration(node, mapper) {
        // If mapper is undefined it means we are checking the declaration of the template.
        if (mapper === undefined) {
            for (const templateParameter of node.templateParameters) {
                checkTemplateParameterDeclaration(templateParameter, undefined);
            }
        }
    }
    function checkModel(node, mapper) {
        if (node.kind === SyntaxKind.ModelStatement) {
            return checkModelStatement(node, mapper);
        }
        else {
            return checkModelExpression(node, mapper);
        }
    }
    function checkModelStatement(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this model and we've already checked it
            return links.declaredType;
        }
        checkTemplateDeclaration(node, mapper);
        const decorators = [];
        const type = createType({
            kind: "Model",
            name: node.id.sv,
            node: node,
            properties: createRekeyableMap(),
            namespace: getParentNamespaceType(node),
            decorators,
            sourceModels: [],
            derivedModels: [],
        });
        linkType(links, type, mapper);
        if (node.symbol.members) {
            const members = resolver.getAugmentedSymbolTable(node.symbol.members);
            const propDocs = extractPropDocs(node);
            for (const [name, memberSym] of members) {
                const doc = propDocs.get(name);
                if (doc) {
                    docFromCommentForSym.set(memberSym, doc);
                }
            }
        }
        const isBase = checkModelIs(node, node.is, mapper);
        if (isBase) {
            type.sourceModel = isBase;
            type.sourceModels.push({ usage: "is", model: isBase });
            // copy decorators
            decorators.push(...isBase.decorators);
            if (isBase.indexer) {
                type.indexer = isBase.indexer;
            }
        }
        if (isBase) {
            for (const prop of isBase.properties.values()) {
                const memberSym = getMemberSymbol(node.symbol, prop.name);
                const newProp = cloneTypeForSymbol(memberSym, prop, {
                    sourceProperty: prop,
                    model: type,
                });
                linkIndirectMember(node, newProp, mapper);
                type.properties.set(prop.name, newProp);
            }
        }
        if (isBase) {
            type.baseModel = isBase.baseModel;
        }
        else if (node.extends) {
            type.baseModel = checkClassHeritage(node, node.extends, mapper);
            if (type.baseModel) {
                copyDeprecation(type.baseModel, type);
            }
        }
        if (type.baseModel) {
            type.baseModel.derivedModels.push(type);
        }
        // Hold on to the model type that's being defined so that it
        // can be referenced
        if (mapper === undefined) {
            type.namespace?.models.set(type.name, type);
        }
        // Evaluate the properties after
        checkModelProperties(node, type.properties, type, mapper);
        decorators.push(...checkDecorators(type, node, mapper));
        linkMapper(type, mapper);
        if (shouldCreateTypeForTemplate(node, mapper)) {
            finishType(type);
        }
        const indexer = getIndexer(program, type);
        if (type.name === "Array" && isInTypeSpecNamespace(type)) {
            stdTypes.Array = type;
        }
        else if (type.name === "Record" && isInTypeSpecNamespace(type)) {
            stdTypes.Record = type;
        }
        if (indexer) {
            type.indexer = indexer;
        }
        lateBindMemberContainer(type);
        lateBindMembers(type);
        return type;
    }
    function shouldCreateTypeForTemplate(node, mapper) {
        // Node is not a template we should create the type.
        if (node.templateParameters.length === 0) {
            return true;
        }
        // There is no mapper so we shouldn't be instantiating the template.
        if (mapper === undefined) {
            return false;
        }
        // Some of the mapper args are still template parameter so we shouldn't create the type.
        return (!mapper.partial &&
            mapper.args.every((t) => isValue(t) || t.entityKind === "Indeterminate" || t.kind !== "TemplateParameter"));
    }
    function checkModelExpression(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this model and we've already checked it
            return links.declaredType;
        }
        const properties = createRekeyableMap();
        const type = createType({
            kind: "Model",
            name: "",
            node: node,
            properties,
            indexer: undefined,
            namespace: getParentNamespaceType(node),
            decorators: [],
            derivedModels: [],
            sourceModels: [],
        });
        linkType(links, type, mapper);
        linkMapper(type, mapper);
        checkModelProperties(node, properties, type, mapper);
        return finishType(type);
    }
    /** Find the indexer that applies to this model. Either defined on itself or from a base model */
    function findIndexer(model) {
        let current = model;
        while (current) {
            if (current.indexer) {
                return current.indexer;
            }
            current = current.baseModel;
        }
        return undefined;
    }
    function checkPropertyCompatibleWithModelIndexer(parentModel, property, diagnosticTarget) {
        const indexer = findIndexer(parentModel);
        if (indexer === undefined) {
            return;
        }
        return checkPropertyCompatibleWithIndexer(indexer, property, diagnosticTarget);
    }
    function checkPropertyCompatibleWithIndexer(indexer, property, diagnosticTarget) {
        if (indexer.key.name === "integer") {
            reportCheckerDiagnostics([
                createDiagnostic({
                    code: "no-array-properties",
                    target: diagnosticTarget,
                }),
            ]);
            return;
        }
        const [valid, diagnostics] = relation.isTypeAssignableTo(property.type, indexer.value, diagnosticTarget);
        if (!valid)
            reportCheckerDiagnostic(createDiagnostic({
                code: "incompatible-indexer",
                format: { message: diagnostics.map((x) => `  ${x.message}`).join("\n") },
                target: diagnosticTarget.kind === SyntaxKind.ModelProperty
                    ? diagnosticTarget.value
                    : diagnosticTarget,
            }));
    }
    function checkModelProperties(node, properties, parentModel, mapper) {
        let spreadIndexers;
        for (const prop of node.properties) {
            if ("id" in prop) {
                const newProp = checkModelProperty(prop, mapper);
                newProp.model = parentModel;
                checkPropertyCompatibleWithModelIndexer(parentModel, newProp, prop);
                defineProperty(properties, newProp);
            }
            else {
                // spread property
                const [newProperties, additionalIndexer] = checkSpreadProperty(node.symbol, prop.target, parentModel, mapper);
                if (additionalIndexer) {
                    if (spreadIndexers) {
                        spreadIndexers.push(additionalIndexer);
                    }
                    else {
                        spreadIndexers = [additionalIndexer];
                    }
                }
                for (const newProp of newProperties) {
                    linkIndirectMember(node, newProp, mapper);
                    checkPropertyCompatibleWithModelIndexer(parentModel, newProp, prop);
                    defineProperty(properties, newProp, prop);
                }
            }
        }
        if (spreadIndexers) {
            const value = spreadIndexers.length === 1
                ? spreadIndexers[0].value
                : createUnion(spreadIndexers.map((i) => i.value));
            parentModel.indexer = {
                key: spreadIndexers[0].key,
                value: value,
            };
        }
    }
    function checkObjectValue(node, mapper, constraint) {
        const properties = checkObjectLiteralProperties(node, mapper);
        if (properties === null) {
            return null;
        }
        const preciseType = createTypeForObjectValue(node, properties);
        if (constraint && !checkTypeOfValueMatchConstraint(preciseType, constraint, node)) {
            return null;
        }
        return createValue({
            entityKind: "Value",
            valueKind: "ObjectValue",
            node: node,
            properties,
            type: constraint ? constraint.type : preciseType,
        }, preciseType);
    }
    function createTypeForObjectValue(node, properties) {
        const model = createType({
            kind: "Model",
            name: "",
            node,
            properties: createRekeyableMap(),
            decorators: [],
            derivedModels: [],
            sourceModels: [],
        });
        for (const prop of properties.values()) {
            model.properties.set(prop.name, createModelPropertyForObjectPropertyDescriptor(prop, model));
        }
        return finishType(model);
    }
    function createModelPropertyForObjectPropertyDescriptor(prop, parentModel) {
        return createAndFinishType({
            kind: "ModelProperty",
            node: prop.node,
            model: parentModel,
            optional: false,
            name: prop.name,
            type: prop.value.type,
            decorators: [],
        });
    }
    function checkObjectLiteralProperties(node, mapper) {
        const properties = new Map();
        let hasError = false;
        for (const prop of node.properties) {
            if ("id" in prop) {
                const value = getValueForNode(prop.value, mapper);
                if (value === null) {
                    hasError = true;
                }
                else {
                    properties.set(prop.id.sv, { name: prop.id.sv, value: value, node: prop });
                }
            }
            else {
                const targetType = checkObjectSpreadProperty(prop.target, mapper);
                if (targetType) {
                    for (const [name, value] of targetType.properties) {
                        properties.set(name, { ...value });
                    }
                }
            }
        }
        return hasError ? null : properties;
    }
    function checkObjectSpreadProperty(targetNode, mapper) {
        const value = getValueForNode(targetNode, mapper);
        if (value === null) {
            return null;
        }
        if (value.valueKind !== "ObjectValue") {
            reportCheckerDiagnostic(createDiagnostic({ code: "spread-object", target: targetNode }));
            return null;
        }
        return value;
    }
    function checkArrayValue(node, mapper, constraint) {
        let hasError = false;
        const values = node.values.map((itemNode) => {
            const value = getValueForNode(itemNode, mapper);
            if (value === null) {
                hasError = true;
            }
            return value;
        });
        if (hasError) {
            return null;
        }
        const preciseType = createTypeForArrayValue(node, values);
        if (constraint && !checkTypeOfValueMatchConstraint(preciseType, constraint, node)) {
            return null;
        }
        return createValue({
            entityKind: "Value",
            valueKind: "ArrayValue",
            node: node,
            values: values,
            type: constraint ? constraint.type : preciseType,
        }, preciseType);
    }
    function createValue(value, preciseType) {
        valueExactTypes.set(value, preciseType);
        return value;
    }
    function copyValue(value, overrides) {
        const newValue = { ...value, ...overrides };
        const preciseType = valueExactTypes.get(value);
        if (preciseType) {
            valueExactTypes.set(newValue, preciseType);
        }
        return newValue;
    }
    function createTypeForArrayValue(node, values) {
        return createAndFinishType({
            kind: "Tuple",
            node,
            values: values.map((x) => x.type),
        });
    }
    function inferScalarForPrimitiveValue(type, literalType) {
        if (type === undefined) {
            return undefined;
        }
        switch (type.kind) {
            case "Scalar":
                if (ignoreDiagnostics(checker.isTypeAssignableTo(literalType, type, literalType))) {
                    return type;
                }
                return undefined;
            case "Union":
                let found = undefined;
                for (const variant of type.variants.values()) {
                    const scalar = inferScalarForPrimitiveValue(variant.type, literalType);
                    if (scalar) {
                        if (found) {
                            reportCheckerDiagnostic(createDiagnostic({
                                code: "ambiguous-scalar-type",
                                format: {
                                    value: getTypeName(literalType),
                                    types: [found, scalar].map((x) => x.name).join(", "),
                                    example: found.name,
                                },
                                target: literalType,
                            }));
                            return undefined;
                        }
                        else {
                            found = scalar;
                        }
                    }
                }
                return found;
            default:
                return undefined;
        }
    }
    function checkStringValue(literalType, constraint, node) {
        if (constraint && !checkTypeOfValueMatchConstraint(literalType, constraint, node)) {
            return null;
        }
        let value;
        if (literalType.kind === "StringTemplate") {
            if (literalType.stringValue) {
                value = literalType.stringValue;
            }
            else {
                reportCheckerDiagnostics(explainStringTemplateNotSerializable(literalType));
                return null;
            }
        }
        else {
            value = literalType.value;
        }
        const scalar = inferScalarForPrimitiveValue(constraint?.type, literalType);
        return createValue({
            entityKind: "Value",
            valueKind: "StringValue",
            value,
            type: constraint ? constraint.type : literalType,
            scalar,
        }, literalType);
    }
    function checkNumericValue(literalType, constraint, node) {
        if (constraint && !checkTypeOfValueMatchConstraint(literalType, constraint, node)) {
            return null;
        }
        const scalar = inferScalarForPrimitiveValue(constraint?.type, literalType);
        return createValue({
            entityKind: "Value",
            valueKind: "NumericValue",
            value: Numeric(literalType.valueAsString),
            type: constraint ? constraint.type : literalType,
            scalar,
        }, literalType);
    }
    function checkBooleanValue(literalType, constraint, node) {
        if (constraint && !checkTypeOfValueMatchConstraint(literalType, constraint, node)) {
            return null;
        }
        const scalar = inferScalarForPrimitiveValue(constraint?.type, literalType);
        return createValue({
            entityKind: "Value",
            valueKind: "BooleanValue",
            value: literalType.value,
            type: constraint ? constraint.type : literalType,
            scalar,
        }, literalType);
    }
    function checkNullValue(literalType, constraint, node) {
        if (constraint && !checkTypeOfValueMatchConstraint(literalType, constraint, node)) {
            return null;
        }
        return createValue({
            entityKind: "Value",
            valueKind: "NullValue",
            type: constraint ? constraint.type : literalType,
            value: null,
        }, literalType);
    }
    function checkEnumValue(literalType, constraint, node) {
        if (constraint && !checkTypeOfValueMatchConstraint(literalType, constraint, node)) {
            return null;
        }
        return createValue({
            entityKind: "Value",
            valueKind: "EnumValue",
            type: constraint ? constraint.type : literalType,
            value: literalType,
        }, literalType);
    }
    function checkCallExpressionTarget(node, mapper) {
        const target = checkTypeReference(node.target, mapper);
        if (target.kind === "Scalar" || target.kind === "ScalarConstructor") {
            return target;
        }
        else {
            reportCheckerDiagnostic(createDiagnostic({
                code: "non-callable",
                format: { type: target.kind },
                target: node.target,
            }));
            return null;
        }
    }
    /** Check the arguments of the call expression are a single value of the given syntax. */
    function checkPrimitiveArg(node, scalar, valueKind) {
        if (node.arguments.length !== 1) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-primitive-init",
                target: node.target,
            }));
            return null;
        }
        const argNode = node.arguments[0];
        const value = getValueForNode(argNode, undefined);
        if (value === null) {
            return null; // error should already have been reported above.
        }
        if (value.valueKind !== valueKind) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-primitive-init",
                messageId: "invalidArg",
                format: { actual: value.valueKind, expected: valueKind },
                target: argNode,
            }));
            return null;
        }
        if (!checkValueOfType(value, scalar, argNode)) {
            return null;
        }
        return copyValue(value, { scalar, type: scalar });
    }
    function createScalarValue(node, mapper, declaration) {
        let hasError = false;
        const minArgs = declaration.parameters.filter((x) => !x.optional && !x.rest).length ?? 0;
        const maxArgs = declaration.parameters[declaration.parameters.length - 1]?.rest
            ? undefined
            : declaration.parameters.length;
        if (node.arguments.length < minArgs ||
            (maxArgs !== undefined && node.arguments.length > maxArgs)) {
            // In the case we have too little args then this decorator is not applicable.
            // If there is too many args then we can still run the decorator as long as the args are valid.
            if (node.arguments.length < minArgs) {
                hasError = true;
            }
            if (maxArgs === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "invalid-argument-count",
                    messageId: "atLeast",
                    format: { actual: node.arguments.length.toString(), expected: minArgs.toString() },
                    target: node,
                }));
            }
            else {
                const expected = minArgs === maxArgs ? minArgs.toString() : `${minArgs}-${maxArgs}`;
                reportCheckerDiagnostic(createDiagnostic({
                    code: "invalid-argument-count",
                    format: { actual: node.arguments.length.toString(), expected },
                    target: node,
                }));
            }
        }
        const resolvedArgs = [];
        for (const [index, parameter] of declaration.parameters.entries()) {
            if (parameter.rest) {
                const restType = getIndexType(parameter.type);
                if (restType) {
                    for (let i = index; i < node.arguments.length; i++) {
                        const argNode = node.arguments[i];
                        if (argNode) {
                            const arg = getValueForNode(argNode, mapper, { kind: "argument", type: restType });
                            if (arg === null) {
                                hasError = true;
                                continue;
                            }
                            if (checkValueOfType(arg, restType, argNode)) {
                                resolvedArgs.push(arg);
                            }
                            else {
                                hasError = true;
                            }
                        }
                    }
                }
                break;
            }
            const argNode = node.arguments[index];
            if (argNode) {
                const arg = getValueForNode(argNode, mapper, {
                    kind: "argument",
                    type: parameter.type,
                });
                if (arg === null) {
                    hasError = true;
                    continue;
                }
                if (checkValueOfType(arg, parameter.type, argNode)) {
                    resolvedArgs.push(arg);
                }
                else {
                    hasError = true;
                }
            }
        }
        if (hasError) {
            return null;
        }
        return {
            entityKind: "Value",
            valueKind: "ScalarValue",
            value: {
                name: declaration.name,
                args: resolvedArgs,
            },
            scalar: declaration.scalar,
            type: declaration.scalar,
        };
    }
    function checkCallExpression(node, mapper) {
        const target = checkCallExpressionTarget(node, mapper);
        if (target === null) {
            return null;
        }
        if (target.kind === "ScalarConstructor") {
            return createScalarValue(node, mapper, target);
        }
        if (relation.areScalarsRelated(target, getStdType("string"))) {
            return checkPrimitiveArg(node, target, "StringValue");
        }
        else if (relation.areScalarsRelated(target, getStdType("numeric"))) {
            return checkPrimitiveArg(node, target, "NumericValue");
        }
        else if (relation.areScalarsRelated(target, getStdType("boolean"))) {
            return checkPrimitiveArg(node, target, "BooleanValue");
        }
        else {
            reportCheckerDiagnostic(createDiagnostic({
                code: "named-init-required",
                format: { typeKind: target.kind },
                target: node.target,
            }));
            return null;
        }
    }
    function checkTypeOfExpression(node, mapper) {
        const entity = checkNode(node.target, mapper, undefined);
        if (entity === null) {
            // Shouldn't need to emit error as we assume null value already emitted error when produced
            return errorType;
        }
        if (entity.entityKind === "Indeterminate") {
            return entity.type;
        }
        if (isType(entity)) {
            if (entity.kind === "TemplateParameter") {
                if (entity.constraint === undefined || entity.constraint.type !== undefined) {
                    // means this template constraint will accept values
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "expect-value",
                        messageId: "templateConstraint",
                        format: { name: getTypeName(entity) },
                        target: node.target,
                    }));
                    return errorType;
                }
                else if (entity.constraint.valueType) {
                    return entity.constraint.valueType;
                }
            }
            reportCheckerDiagnostic(createDiagnostic({
                code: "expect-value",
                format: { name: getTypeName(entity) },
                target: node.target,
            }));
            return entity;
        }
        return entity.type;
    }
    function createUnion(options) {
        const variants = createRekeyableMap();
        const union = createAndFinishType({
            kind: "Union",
            node: undefined,
            options,
            decorators: [],
            variants,
            expression: true,
        });
        for (const option of options) {
            const name = Symbol("indexer-union-variant");
            variants.set(name, createAndFinishType({
                kind: "UnionVariant",
                node: undefined,
                type: option,
                name,
                union,
                decorators: [],
            }));
        }
        return union;
    }
    function defineProperty(properties, newProp, diagnosticTarget) {
        if (properties.has(newProp.name)) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "duplicate-property",
                format: { propName: newProp.name },
                target: diagnosticTarget ?? newProp,
            }));
            return;
        }
        const overriddenProp = getOverriddenProperty(newProp);
        if (overriddenProp) {
            const [isAssignable, _] = relation.isTypeAssignableTo(newProp.type, overriddenProp.type, newProp);
            const parentType = getTypeName(overriddenProp.type);
            const newPropType = getTypeName(newProp.type);
            let invalid = false;
            if (!isAssignable) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "override-property-mismatch",
                    format: { propName: newProp.name, propType: newPropType, parentType: parentType },
                    target: diagnosticTarget ?? newProp,
                }));
                invalid = true;
            }
            if (!overriddenProp.optional && newProp.optional) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "override-property-mismatch",
                    messageId: "disallowedOptionalOverride",
                    format: { propName: overriddenProp.name },
                    target: diagnosticTarget ?? newProp,
                }));
                invalid = true;
            }
            if (invalid)
                return;
        }
        properties.set(newProp.name, newProp);
    }
    /**
     * Initializes a late bound symbol for the type. This is generally necessary when attempting to
     * access a symbol for a type that is created during the check phase.
     */
    function lateBindMemberContainer(type) {
        if (type.symbol)
            return;
        switch (type.kind) {
            case "Model":
                type.symbol = createSymbol(type.node, type.name, 2 /* SymbolFlags.Model */ | 4194304 /* SymbolFlags.LateBound */);
                mutate(type.symbol).type = type;
                break;
            case "Interface":
                type.symbol = createSymbol(type.node, type.name, 32 /* SymbolFlags.Interface */ | 4194304 /* SymbolFlags.LateBound */);
                mutate(type.symbol).type = type;
                break;
            case "Union":
                if (!type.name)
                    return; // don't make a symbol for anonymous unions
                type.symbol = createSymbol(type.node, type.name, 64 /* SymbolFlags.Union */ | 4194304 /* SymbolFlags.LateBound */);
                mutate(type.symbol).type = type;
                break;
        }
    }
    function lateBindMembers(type) {
        compilerAssert(type.symbol, "Type must have a symbol to late bind members");
        const containerSym = type.symbol;
        compilerAssert(containerSym.members, "Container symbol didn't have members at late-bind");
        const containerMembers = resolver.getAugmentedSymbolTable(containerSym.members);
        switch (type.kind) {
            case "Model":
                for (const prop of walkPropertiesInherited(type)) {
                    lateBindMember(prop, 65536 /* SymbolFlags.Member */ | 1048576 /* SymbolFlags.Declaration */);
                }
                break;
            case "Scalar":
                for (const member of type.constructors.values()) {
                    lateBindMember(member, 65536 /* SymbolFlags.Member */ | 1048576 /* SymbolFlags.Declaration */);
                }
                break;
            case "Enum":
                for (const member of type.members.values()) {
                    lateBindMember(member, 65536 /* SymbolFlags.Member */ | 1048576 /* SymbolFlags.Declaration */);
                }
                break;
            case "Interface":
                for (const member of type.operations.values()) {
                    lateBindMember(member, 65536 /* SymbolFlags.Member */ | 8 /* SymbolFlags.Operation */ | 1048576 /* SymbolFlags.Declaration */);
                }
                break;
            case "Union":
                for (const variant of type.variants.values()) {
                    lateBindMember(variant, 65536 /* SymbolFlags.Member */ | 1048576 /* SymbolFlags.Declaration */);
                }
                break;
        }
        function lateBindMember(member, kind) {
            if (!member.node || typeof member.name !== "string") {
                // don't bind anything for union expressions
                return;
            }
            const sym = createSymbol(member.node, member.name, kind | 4194304 /* SymbolFlags.LateBound */, containerSym);
            mutate(sym).type = member;
            compilerAssert(containerSym.members, "containerSym.members is undefined");
            containerMembers.set(member.name, sym);
        }
    }
    function checkClassHeritage(model, heritageRef, mapper) {
        if (heritageRef.kind === SyntaxKind.ModelExpression) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "extend-model",
                messageId: "modelExpression",
                target: heritageRef,
            }));
            return undefined;
        }
        if (heritageRef.kind !== SyntaxKind.TypeReference &&
            heritageRef.kind !== SyntaxKind.ArrayExpression) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "extend-model",
                target: heritageRef,
            }));
            return undefined;
        }
        const modelSymId = getNodeSym(model);
        pendingResolutions.start(modelSymId, ResolutionKind.BaseType);
        const target = resolver.getNodeLinks(heritageRef).resolvedSymbol;
        if (target && pendingResolutions.has(target, ResolutionKind.BaseType)) {
            if (mapper === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "circular-base-type",
                    format: { typeName: target.name },
                    target: target,
                }));
            }
            return undefined;
        }
        const heritageType = getTypeForNode(heritageRef, mapper);
        pendingResolutions.finish(modelSymId, ResolutionKind.BaseType);
        if (isErrorType(heritageType)) {
            compilerAssert(program.hasError(), "Should already have reported an error.", heritageRef);
            return undefined;
        }
        if (heritageType.kind !== "Model") {
            reportCheckerDiagnostic(createDiagnostic({ code: "extend-model", target: heritageRef }));
            return undefined;
        }
        if (heritageType.name === "") {
            reportCheckerDiagnostic(createDiagnostic({
                code: "extend-model",
                messageId: "modelExpression",
                target: heritageRef,
            }));
        }
        return heritageType;
    }
    function checkModelIs(model, isExpr, mapper) {
        if (!isExpr)
            return undefined;
        const modelSymId = getNodeSym(model);
        pendingResolutions.start(modelSymId, ResolutionKind.BaseType);
        let isType;
        if (isExpr.kind === SyntaxKind.ModelExpression) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "is-model",
                messageId: "modelExpression",
                target: isExpr,
            }));
            return undefined;
        }
        else if (isExpr.kind === SyntaxKind.ArrayExpression) {
            isType = checkArrayExpression(isExpr, mapper);
        }
        else if (isExpr.kind === SyntaxKind.TypeReference) {
            const target = resolver.getNodeLinks(isExpr).resolvedSymbol;
            if (target && pendingResolutions.has(target, ResolutionKind.BaseType)) {
                if (mapper === undefined) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "circular-base-type",
                        format: { typeName: target.name },
                        target: target,
                    }));
                }
                return undefined;
            }
            isType = getTypeForNode(isExpr, mapper);
        }
        else {
            reportCheckerDiagnostic(createDiagnostic({ code: "is-model", target: isExpr }));
            return undefined;
        }
        pendingResolutions.finish(modelSymId, ResolutionKind.BaseType);
        if (isType.kind !== "Model") {
            reportCheckerDiagnostic(createDiagnostic({ code: "is-model", target: isExpr }));
            return;
        }
        if (isType.name === "") {
            reportCheckerDiagnostic(createDiagnostic({ code: "is-model", messageId: "modelExpression", target: isExpr }));
            return undefined;
        }
        return isType;
    }
    function checkSpreadProperty(parentModelSym, targetNode, parentModel, mapper) {
        const targetType = getTypeForNode(targetNode, mapper);
        if (targetType.kind === "TemplateParameter" || isErrorType(targetType)) {
            return [[], undefined];
        }
        if (targetType.kind !== "Model") {
            reportCheckerDiagnostic(createDiagnostic({ code: "spread-model", target: targetNode }), mapper);
            return [[], undefined];
        }
        if (isArrayModelType(program, targetType)) {
            reportCheckerDiagnostic(createDiagnostic({ code: "spread-model", target: targetNode }), mapper);
            return [[], undefined];
        }
        if (parentModel === targetType) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "spread-model",
                messageId: "selfSpread",
                target: targetNode,
            }));
        }
        parentModel.sourceModels.push({ usage: "spread", model: targetType });
        const props = [];
        // copy each property
        for (const prop of walkPropertiesInherited(targetType)) {
            const memberSym = getMemberSymbol(parentModelSym, prop.name);
            props.push(cloneTypeForSymbol(memberSym, prop, {
                sourceProperty: prop,
                model: parentModel,
            }));
        }
        return [props, targetType.indexer];
    }
    /**
     * Link an indirect model property(included via spread or model is) to its model member symbols.
     * @param containerNode Model Node
     * @param member New Property
     * @param mapper Type Mapper.
     */
    function linkIndirectMember(containerNode, member, mapper) {
        if (mapper !== undefined) {
            return;
        }
        compilerAssert(typeof member.name === "string", "Cannot link unmapped unions");
        if (containerNode.symbol === undefined) {
            return;
        }
        const memberSym = getMemberSymbol(containerNode.symbol, member.name);
        if (memberSym) {
            const links = resolver.getSymbolLinks(memberSym);
            linkMemberType(links, member, mapper);
        }
    }
    function checkModelProperty(prop, mapper) {
        const sym = getSymbolForMember(prop);
        const links = getSymbolLinksForMember(prop);
        if (links && links.declaredType && mapper === undefined) {
            return links.declaredType;
        }
        const name = prop.id.sv;
        const type = createType({
            kind: "ModelProperty",
            name,
            node: prop,
            optional: prop.optional,
            type: undefined,
            decorators: [],
        });
        if (pendingResolutions.has(sym, ResolutionKind.Type) && mapper === undefined) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "circular-prop",
                format: { propName: name },
                target: prop,
            }));
            type.type = errorType;
        }
        else {
            pendingResolutions.start(sym, ResolutionKind.Type);
            type.type = getTypeForNode(prop.value, mapper);
            if (prop.default) {
                const defaultValue = checkDefaultValue(prop.default, type.type, mapper);
                if (defaultValue !== null) {
                    type.defaultValue = defaultValue;
                }
            }
            if (links) {
                linkType(links, type, mapper);
            }
        }
        type.decorators = checkDecorators(type, prop, mapper);
        const parentTemplate = getParentTemplateNode(prop);
        linkMapper(type, mapper);
        if (!parentTemplate || shouldCreateTypeForTemplate(parentTemplate, mapper)) {
            const docComment = docFromCommentForSym.get(sym);
            if (docComment) {
                type.decorators.unshift(createDocFromCommentDecorator("self", docComment));
            }
            finishType(type);
        }
        pendingResolutions.finish(sym, ResolutionKind.Type);
        return type;
    }
    function createDocFromCommentDecorator(key, doc) {
        return {
            decorator: docFromCommentDecorator,
            args: [
                { value: createLiteralType(key), jsValue: key },
                { value: createLiteralType(doc), jsValue: doc },
            ],
        };
    }
    function checkDefaultValue(defaultNode, type, mapper) {
        if (isErrorType(type)) {
            // if the prop type is an error we don't need to validate again.
            return null;
        }
        const defaultValue = getValueForNode(defaultNode, mapper, {
            kind: "assignment",
            type,
        });
        if (defaultValue === null) {
            return null;
        }
        const [related, diagnostics] = relation.isValueOfType(defaultValue, type, defaultNode);
        if (!related) {
            reportCheckerDiagnostics(diagnostics);
            return null;
        }
        else if (defaultValue.valueKind === "TemplateValue") {
            // Right now we don't want to expose `TemplateValue` in the type graph.
            // And as interating with the template declaration is not a supported feature we can just drop it.
            return null;
        }
        else {
            return { ...defaultValue, type };
        }
    }
    function checkDecoratorApplication(targetType, decNode, mapper) {
        const sym = resolveTypeReferenceSym(decNode.target, undefined, true);
        if (!sym) {
            // Error should already have been reported above
            return undefined;
        }
        if (!(sym.flags & 512 /* SymbolFlags.Decorator */)) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-decorator",
                format: { id: sym.name },
                target: decNode,
            }));
            return undefined;
        }
        const symbolLinks = getSymbolLinks(sym);
        let hasError = false;
        if (symbolLinks.declaredType === undefined) {
            const decoratorDeclNode = sym.declarations.find((x) => x.kind === SyntaxKind.DecoratorDeclarationStatement);
            if (decoratorDeclNode) {
                checkDecoratorDeclaration(decoratorDeclNode, undefined);
            }
        }
        if (symbolLinks.declaredType) {
            compilerAssert(symbolLinks.declaredType.kind === "Decorator", `Expected to find a decorator type but got ${symbolLinks.declaredType.kind}`);
            if (!checkDecoratorTarget(targetType, symbolLinks.declaredType, decNode)) {
                hasError = true;
            }
        }
        const [argsHaveError, args] = checkDecoratorArguments(decNode, mapper, symbolLinks.declaredType);
        if (hasError || argsHaveError) {
            return undefined;
        }
        return {
            definition: symbolLinks.declaredType,
            decorator: sym.value ?? ((...args) => { }),
            node: decNode,
            args,
        };
    }
    /** Check the decorator target is valid */
    function checkDecoratorTarget(targetType, declaration, decoratorNode) {
        const [targetValid] = relation.isTypeAssignableTo(targetType, declaration.target.type, decoratorNode);
        if (!targetValid) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "decorator-wrong-target",
                messageId: "withExpected",
                format: {
                    decorator: declaration.name,
                    to: getTypeName(targetType),
                    expected: getEntityName(declaration.target.type),
                },
                target: decoratorNode,
            }));
        }
        return targetValid;
    }
    function checkDecoratorArguments(node, mapper, declaration) {
        // if we don't have a declaration we can just return the types or values if
        if (declaration === undefined) {
            return [
                false,
                node.arguments.map((argNode) => {
                    let type = checkNode(argNode, mapper) ?? errorType;
                    if (type.entityKind === "Indeterminate") {
                        type = type.type;
                    }
                    return {
                        value: type,
                        jsValue: type,
                        node: argNode,
                    };
                }),
            ];
        }
        let hasError = false;
        const minArgs = declaration.parameters.filter((x) => !x.optional && !x.rest).length ?? 0;
        const maxArgs = declaration.parameters[declaration.parameters.length - 1]?.rest
            ? undefined
            : declaration.parameters.length;
        if (node.arguments.length < minArgs ||
            (maxArgs !== undefined && node.arguments.length > maxArgs)) {
            // In the case we have too little args then this decorator is not applicable.
            // If there is too many args then we can still run the decorator as long as the args are valid.
            if (node.arguments.length < minArgs) {
                hasError = true;
            }
            if (maxArgs === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "invalid-argument-count",
                    messageId: "atLeast",
                    format: { actual: node.arguments.length.toString(), expected: minArgs.toString() },
                    target: node,
                }));
            }
            else {
                const expected = minArgs === maxArgs ? minArgs.toString() : `${minArgs}-${maxArgs}`;
                reportCheckerDiagnostic(createDiagnostic({
                    code: "invalid-argument-count",
                    format: { actual: node.arguments.length.toString(), expected },
                    target: node,
                }));
            }
        }
        const resolvedArgs = [];
        function resolveArg(argNode, perParamType) {
            const arg = getTypeOrValueForNode(argNode, mapper, {
                kind: "argument",
                constraint: perParamType,
            });
            if (arg !== null &&
                !(isType(arg) && isErrorType(arg)) &&
                checkArgumentAssignable(arg, perParamType, argNode)) {
                return {
                    value: arg,
                    node: argNode,
                    jsValue: resolveDecoratorArgJsValue(arg, extractValueOfConstraints({
                        kind: "argument",
                        constraint: perParamType,
                    })),
                };
            }
            else {
                return undefined;
            }
        }
        for (const [index, parameter] of declaration.parameters.entries()) {
            if (parameter.rest) {
                const restType = extractRestParamConstraint(parameter.type);
                if (restType) {
                    for (let i = index; i < node.arguments.length; i++) {
                        const argNode = node.arguments[i];
                        if (argNode) {
                            const arg = resolveArg(argNode, restType);
                            if (arg) {
                                resolvedArgs.push(arg);
                            }
                            else {
                                hasError = true;
                            }
                        }
                    }
                }
                break;
            }
            const argNode = node.arguments[index];
            if (argNode) {
                const arg = resolveArg(argNode, parameter.type);
                if (arg) {
                    resolvedArgs.push(arg);
                }
                else {
                    hasError = true;
                }
            }
        }
        return [hasError, resolvedArgs];
    }
    /** For a rest param of constraint T[] or valueof T[] return the T or valueof T */
    function extractRestParamConstraint(constraint) {
        let valueType;
        let type;
        if (constraint.valueType) {
            if (constraint.valueType.kind === "Model" &&
                isArrayModelType(program, constraint.valueType)) {
                valueType = constraint.valueType.indexer.value;
            }
            else {
                return undefined;
            }
        }
        if (constraint.type) {
            if (constraint.type.kind === "Model" && isArrayModelType(program, constraint.type)) {
                type = constraint.type.indexer.value;
            }
            else {
                return undefined;
            }
        }
        return {
            entityKind: "MixedParameterConstraint",
            type,
            valueType,
        };
    }
    function getIndexType(type) {
        return type.kind === "Model" ? type.indexer?.value : undefined;
    }
    function resolveDecoratorArgJsValue(value, valueConstraint) {
        if (valueConstraint !== undefined) {
            if (isValue(value)) {
                return marshallTypeForJS(value, valueConstraint.type);
            }
            else {
                return value;
            }
        }
        return value;
    }
    function checkArgumentAssignable(argumentType, parameterType, diagnosticTarget) {
        const [valid] = relation.isTypeAssignableTo(argumentType, parameterType, diagnosticTarget);
        if (!valid) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "invalid-argument",
                format: {
                    value: getEntityName(argumentType),
                    expected: getEntityName(parameterType),
                },
                target: diagnosticTarget,
            }));
        }
        return valid;
    }
    function checkAugmentDecorators(sym, targetType, mapper) {
        const augmentDecoratorNodes = resolver.getAugmentDecoratorsForSym(sym);
        const decorators = [];
        for (const decNode of augmentDecoratorNodes) {
            const decorator = checkDecoratorApplication(targetType, decNode, mapper);
            if (decorator) {
                decorators.unshift(decorator);
            }
        }
        return decorators;
    }
    /**
     * Check that augment decorator are targeting valid symbols.
     */
    function checkAugmentDecorator(node) {
        // This will validate the target type is pointing to a valid ref.
        resolveTypeReferenceSym(node.targetType, undefined);
        const links = resolver.getNodeLinks(node.targetType);
        if (links.isTemplateInstantiation) {
            program.reportDiagnostic(createDiagnostic({
                code: "augment-decorator-target",
                messageId: "noInstance",
                target: node.targetType,
            }));
        }
        else if (links.finalSymbol?.flags &&
            ~links.finalSymbol.flags & 1048576 /* SymbolFlags.Declaration */ &&
            ~links.finalSymbol.flags & 65536 /* SymbolFlags.Member */) {
            program.reportDiagnostic(createDiagnostic({
                code: "augment-decorator-target",
                messageId: links.finalSymbol.flags & 2 /* SymbolFlags.Model */
                    ? "noModelExpression"
                    : links.finalSymbol.flags & 64 /* SymbolFlags.Union */
                        ? "noUnionExpression"
                        : "default",
                target: node.targetType,
            }));
        }
        else if (links.finalSymbol?.flags && links.finalSymbol.flags & 128 /* SymbolFlags.Alias */) {
            const aliasNode = getSymNode(links.finalSymbol);
            program.reportDiagnostic(createDiagnostic({
                code: "augment-decorator-target",
                messageId: aliasNode.value.kind === SyntaxKind.UnionExpression ? "noUnionExpression" : "default",
                target: node.targetType,
            }));
        }
        // If this was used to get a type this is invalid, only used for validation.
        return errorType;
    }
    /**
     * Check that using statements are targeting valid symbols.
     */
    function checkUsings(node) {
        const usedSym = resolveTypeReferenceSym(node.name, undefined);
        if (usedSym) {
            if (~usedSym.flags & 256 /* SymbolFlags.Namespace */) {
                reportCheckerDiagnostic(createDiagnostic({ code: "using-invalid-ref", target: node.name }));
            }
        }
        // If this was used to get a type this is invalid, only used for validation.
        return errorType;
    }
    function checkDecorators(targetType, node, mapper) {
        const sym = isMemberNode(node)
            ? (getSymbolForMember(node) ?? node.symbol)
            : getMergedSymbol(node.symbol);
        const decorators = [];
        const augmentDecoratorNodes = resolver.getAugmentDecoratorsForSym(sym);
        const decoratorNodes = [
            ...augmentDecoratorNodes, // the first decorator will be executed at last, so augmented decorator should be placed at first.
            ...node.decorators,
        ];
        for (const decNode of decoratorNodes) {
            const decorator = checkDecoratorApplication(targetType, decNode, mapper);
            if (decorator) {
                decorators.unshift(decorator);
            }
        }
        // Doc comment should always be the first decorator in case an explicit @doc must override it.
        const docComment = extractMainDoc(targetType);
        if (docComment) {
            decorators.unshift(createDocFromCommentDecorator("self", docComment));
        }
        if (targetType.kind === "Operation") {
            const returnTypesDocs = extractReturnsDocs(targetType);
            if (returnTypesDocs.returns) {
                decorators.unshift(createDocFromCommentDecorator("returns", returnTypesDocs.returns));
            }
            if (returnTypesDocs.errors) {
                decorators.unshift(createDocFromCommentDecorator("errors", returnTypesDocs.errors));
            }
        }
        else if (targetType.kind === "ModelProperty") {
        }
        return decorators;
    }
    function checkScalar(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this model and we've already checked it
            return links.declaredType;
        }
        checkTemplateDeclaration(node, mapper);
        const decorators = [];
        const type = createType({
            kind: "Scalar",
            name: node.id.sv,
            node: node,
            constructors: new Map(),
            namespace: getParentNamespaceType(node),
            decorators,
            derivedScalars: [],
        });
        linkType(links, type, mapper);
        if (node.extends) {
            type.baseScalar = checkScalarExtends(node, node.extends, mapper);
            if (type.baseScalar) {
                copyDeprecation(type.baseScalar, type);
                type.baseScalar.derivedScalars.push(type);
            }
        }
        checkScalarConstructors(type, node, type.constructors, mapper);
        decorators.push(...checkDecorators(type, node, mapper));
        if (mapper === undefined) {
            type.namespace?.scalars.set(type.name, type);
        }
        linkMapper(type, mapper);
        if (shouldCreateTypeForTemplate(node, mapper)) {
            finishType(type);
        }
        if (isInTypeSpecNamespace(type)) {
            stdTypes[type.name] = type;
        }
        return type;
    }
    function checkScalarExtends(scalar, extendsRef, mapper) {
        const symId = getNodeSym(scalar);
        pendingResolutions.start(symId, ResolutionKind.BaseType);
        const target = resolver.getNodeLinks(extendsRef).resolvedSymbol;
        if (target && pendingResolutions.has(target, ResolutionKind.BaseType)) {
            if (mapper === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "circular-base-type",
                    format: { typeName: target.declarations[0].id.sv },
                    target: target,
                }));
            }
            return undefined;
        }
        const extendsType = getTypeForNode(extendsRef, mapper);
        pendingResolutions.finish(symId, ResolutionKind.BaseType);
        if (isErrorType(extendsType)) {
            compilerAssert(program.hasError(), "Should already have reported an error.", extendsRef);
            return undefined;
        }
        if (extendsType.kind !== "Scalar") {
            reportCheckerDiagnostic(createDiagnostic({ code: "extend-model", target: extendsRef }));
            return undefined;
        }
        return extendsType;
    }
    function checkScalarConstructors(parentScalar, node, constructors, mapper) {
        if (parentScalar.baseScalar) {
            for (const member of parentScalar.baseScalar.constructors.values()) {
                const newConstructor = cloneTypeForSymbol(getMemberSymbol(node.symbol, member.name), {
                    ...member,
                    scalar: parentScalar,
                });
                linkIndirectMember(node, newConstructor, mapper);
                constructors.set(member.name, newConstructor);
            }
        }
        for (const member of node.members) {
            const constructor = checkScalarConstructor(member, mapper, parentScalar);
            if (constructors.has(constructor.name)) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "constructor-duplicate",
                    format: { name: constructor.name.toString() },
                    target: member,
                }));
                continue;
            }
            constructors.set(constructor.name, constructor);
        }
    }
    function checkScalarConstructor(node, mapper, parentScalar) {
        const name = node.id.sv;
        const links = getSymbolLinksForMember(node);
        if (links && links.declaredType && mapper === undefined) {
            // we're not instantiating this scalar constructor and we've already checked it
            return links.declaredType;
        }
        const member = createType({
            kind: "ScalarConstructor",
            scalar: parentScalar,
            name,
            node,
            parameters: node.parameters.map((x) => checkFunctionParameter(x, mapper, false)),
        });
        linkMapper(member, mapper);
        if (shouldCreateTypeForTemplate(node.parent, mapper)) {
            finishType(member);
        }
        if (links) {
            linkType(links, member, mapper);
        }
        return finishType(member);
    }
    function checkAlias(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            return links.declaredType;
        }
        checkTemplateDeclaration(node, mapper);
        const aliasSymId = getNodeSym(node);
        if (pendingResolutions.has(aliasSymId, ResolutionKind.Type)) {
            if (mapper === undefined) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "circular-alias-type",
                    format: { typeName: node.id.sv },
                    target: node,
                }));
            }
            links.declaredType = errorType;
            return errorType;
        }
        pendingResolutions.start(aliasSymId, ResolutionKind.Type);
        const type = checkNode(node.value, mapper);
        if (type === null) {
            links.declaredType = errorType;
            return errorType;
        }
        if (isValue(type)) {
            reportCheckerDiagnostic(createDiagnostic({ code: "value-in-type", target: node.value }));
            links.declaredType = errorType;
            return errorType;
        }
        linkType(links, type, mapper);
        pendingResolutions.finish(aliasSymId, ResolutionKind.Type);
        return type;
    }
    function checkConst(node) {
        const links = getSymbolLinks(node.symbol);
        if (links.value !== undefined) {
            return links.value;
        }
        const type = node.type ? getTypeForNode(node.type, undefined) : undefined;
        if (pendingResolutions.has(node.symbol, ResolutionKind.Value)) {
            reportCheckerDiagnostic(createDiagnostic({
                code: "circular-const",
                format: { name: node.id.sv },
                target: node,
            }));
            return null;
        }
        pendingResolutions.start(node.symbol, ResolutionKind.Value);
        const value = getValueForNode(node.value, undefined, type && { kind: "assignment", type });
        pendingResolutions.finish(node.symbol, ResolutionKind.Value);
        if (value === null || (type && !checkValueOfType(value, type, node.id))) {
            links.value = null;
            return links.value;
        }
        links.value = type ? copyValue(value, { type }) : copyValue(value);
        return links.value;
    }
    function inferScalarsFromConstraints(value, type) {
        switch (value.valueKind) {
            case "BooleanValue":
            case "StringValue":
            case "NumericValue":
                if (value.scalar === undefined) {
                    const scalar = inferScalarForPrimitiveValue(type, value.type);
                    return copyValue(value, { scalar });
                }
                return value;
            case "ArrayValue":
            case "ObjectValue":
            case "EnumValue":
            case "NullValue":
            case "ScalarValue":
                return value;
        }
    }
    function checkEnum(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (!links.type) {
            const enumType = (links.type = createType({
                kind: "Enum",
                name: node.id.sv,
                node,
                members: createRekeyableMap(),
                decorators: [],
            }));
            const memberNames = new Set();
            for (const member of node.members) {
                if (member.kind === SyntaxKind.EnumMember) {
                    const memberType = checkEnumMember(member, mapper, enumType);
                    if (memberNames.has(memberType.name)) {
                        reportCheckerDiagnostic(createDiagnostic({
                            code: "enum-member-duplicate",
                            format: { name: memberType.name },
                            target: node,
                        }));
                        continue;
                    }
                    memberNames.add(memberType.name);
                    enumType.members.set(memberType.name, memberType);
                }
                else {
                    const members = checkEnumSpreadMember(node.symbol, enumType, member.target, mapper, memberNames);
                    for (const memberType of members) {
                        linkIndirectMember(node, memberType, mapper);
                        enumType.members.set(memberType.name, memberType);
                    }
                }
            }
            const namespace = getParentNamespaceType(node);
            enumType.namespace = namespace;
            enumType.namespace?.enums.set(enumType.name, enumType);
            enumType.decorators = checkDecorators(enumType, node, mapper);
            linkMapper(enumType, mapper);
            finishType(enumType);
        }
        return links.type;
    }
    function checkInterface(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this interface and we've already checked it
            return links.declaredType;
        }
        checkTemplateDeclaration(node, mapper);
        const interfaceType = createType({
            kind: "Interface",
            decorators: [],
            node,
            namespace: getParentNamespaceType(node),
            sourceInterfaces: [],
            operations: createRekeyableMap(),
            name: node.id.sv,
        });
        linkType(links, interfaceType, mapper);
        interfaceType.decorators = checkDecorators(interfaceType, node, mapper);
        const ownMembers = checkInterfaceMembers(node, mapper, interfaceType);
        for (const extendsNode of node.extends) {
            const extendsType = getTypeForNode(extendsNode, mapper);
            if (extendsType.kind !== "Interface") {
                reportCheckerDiagnostic(createDiagnostic({ code: "extends-interface", target: extendsNode }));
                continue;
            }
            for (const member of extendsType.operations.values()) {
                if (interfaceType.operations.has(member.name)) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "extends-interface-duplicate",
                        format: { name: member.name },
                        target: extendsNode,
                    }));
                }
                const newMember = cloneTypeForSymbol(getMemberSymbol(node.symbol, member.name), member, {
                    interface: interfaceType,
                });
                // Don't link it it is overritten
                if (!ownMembers.has(member.name)) {
                    linkIndirectMember(node, newMember, mapper);
                }
                // Clone deprecation information
                copyDeprecation(member, newMember);
                interfaceType.operations.set(newMember.name, newMember);
            }
            interfaceType.sourceInterfaces.push(extendsType);
        }
        for (const [key, value] of ownMembers) {
            interfaceType.operations.set(key, value);
        }
        linkMapper(interfaceType, mapper);
        if (shouldCreateTypeForTemplate(node, mapper)) {
            finishType(interfaceType);
        }
        if (mapper === undefined) {
            interfaceType.namespace?.interfaces.set(interfaceType.name, interfaceType);
        }
        lateBindMemberContainer(interfaceType);
        lateBindMembers(interfaceType);
        return interfaceType;
    }
    function checkInterfaceMembers(node, mapper, interfaceType) {
        const ownMembers = new Map();
        for (const opNode of node.operations) {
            const opType = checkOperation(opNode, mapper, interfaceType);
            if (ownMembers.has(opType.name)) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "interface-duplicate",
                    format: { name: opType.name },
                    target: opNode,
                }));
                continue;
            }
            ownMembers.set(opType.name, opType);
        }
        return ownMembers;
    }
    function checkUnion(node, mapper) {
        const links = getSymbolLinks(node.symbol);
        if (links.declaredType && mapper === undefined) {
            // we're not instantiating this union and we've already checked it
            return links.declaredType;
        }
        checkTemplateDeclaration(node, mapper);
        const variants = createRekeyableMap();
        const unionType = createType({
            kind: "Union",
            decorators: [],
            node,
            namespace: getParentNamespaceType(node),
            name: node.id.sv,
            variants,
            get options() {
                return Array.from(this.variants.values()).map((v) => v.type);
            },
            expression: false,
        });
        linkType(links, unionType, mapper);
        unionType.decorators = checkDecorators(unionType, node, mapper);
        checkUnionVariants(unionType, node, variants, mapper);
        linkMapper(unionType, mapper);
        if (shouldCreateTypeForTemplate(node, mapper)) {
            finishType(unionType);
        }
        if (mapper === undefined) {
            unionType.namespace?.unions.set(unionType.name, unionType);
        }
        lateBindMemberContainer(unionType);
        lateBindMembers(unionType);
        return unionType;
    }
    function checkUnionVariants(parentUnion, node, variants, mapper) {
        for (const variantNode of node.options) {
            const variantType = checkUnionVariant(variantNode, mapper);
            variantType.union = parentUnion;
            if (variants.has(variantType.name)) {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "union-duplicate",
                    format: { name: variantType.name.toString() },
                    target: variantNode,
                }));
                continue;
            }
            variants.set(variantType.name, variantType);
        }
    }
    function checkUnionVariant(variantNode, mapper) {
        const links = getSymbolLinksForMember(variantNode);
        if (links && links.declaredType && mapper === undefined) {
            // we're not instantiating this union variant and we've already checked it
            return links.declaredType;
        }
        const name = variantNode.id ? variantNode.id.sv : Symbol("name");
        const type = getTypeForNode(variantNode.value, mapper);
        const variantType = createType({
            kind: "UnionVariant",
            name,
            node: variantNode,
            decorators: [],
            type,
            union: undefined,
        });
        variantType.decorators = checkDecorators(variantType, variantNode, mapper);
        linkMapper(variantType, mapper);
        if (shouldCreateTypeForTemplate(variantNode.parent, mapper)) {
            finishType(variantType);
        }
        if (links) {
            linkType(links, variantType, mapper);
        }
        return variantType;
    }
    function isMemberNode(node) {
        return node.symbol && !!(node.symbol.flags & 65536 /* SymbolFlags.Member */);
    }
    function getMemberSymbol(parentSym, name) {
        return parentSym ? resolver.getAugmentedSymbolTable(parentSym.members).get(name) : undefined;
    }
    function getSymbolForMember(node) {
        if (!node.id) {
            return undefined;
        }
        const name = node.id.sv;
        const parentSym = node.parent?.symbol;
        return parentSym ? getMemberSymbol(parentSym, name) : undefined;
    }
    function getSymbolLinksForMember(node) {
        const sym = getSymbolForMember(node);
        return sym ? (getSymNode(sym) === node ? getSymbolLinks(sym) : undefined) : undefined;
    }
    function checkEnumMember(node, mapper, parentEnum) {
        const name = node.id.sv;
        const links = getSymbolLinksForMember(node);
        if (links?.type) {
            return links.type;
        }
        compilerAssert(parentEnum, "Enum member should already have been checked.");
        const value = node.value ? node.value.value : undefined;
        const member = createType({
            kind: "EnumMember",
            enum: parentEnum,
            name,
            node,
            value,
            decorators: [],
        });
        if (links) {
            links.type = member;
        }
        member.decorators = checkDecorators(member, node, mapper);
        return finishType(member);
    }
    function checkEnumSpreadMember(parentEnumSym, parentEnum, targetNode, mapper, existingMemberNames) {
        const members = [];
        const targetType = getTypeForNode(targetNode, mapper);
        if (!isErrorType(targetType)) {
            if (targetType.kind !== "Enum") {
                reportCheckerDiagnostic(createDiagnostic({ code: "spread-enum", target: targetNode }));
                return members;
            }
            for (const member of targetType.members.values()) {
                if (existingMemberNames.has(member.name)) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "enum-member-duplicate",
                        format: { name: member.name },
                        target: targetNode,
                    }));
                }
                else {
                    existingMemberNames.add(member.name);
                    const memberSym = getMemberSymbol(parentEnumSym, member.name);
                    const clonedMember = cloneTypeForSymbol(memberSym, member, {
                        enum: parentEnum,
                        sourceMember: member,
                    });
                    if (clonedMember) {
                        members.push(clonedMember);
                    }
                }
            }
        }
        return members;
    }
    function checkDirectives(node, type) {
        let hasDeprecation = false;
        for (const directive of node.directives ?? []) {
            if (directive.target.sv === "deprecated") {
                const message = directive.arguments[0];
                if (message === undefined) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-deprecation-argument",
                        messageId: "missing",
                        target: directive,
                    }));
                    continue;
                }
                let messageStr;
                if (message.kind !== SyntaxKind.StringLiteral) {
                    reportCheckerDiagnostic(createDiagnostic({
                        code: "invalid-deprecation-argument",
                        format: { kind: SyntaxKind[message.kind] },
                        target: directive.arguments[0],
                    }));
                    messageStr = "<missing message>";
                }
                else {
                    messageStr = message.value;
                }
                if (hasDeprecation === true) {
                    reportCheckerDiagnostic(createDiagnostic({ code: "duplicate-deprecation", target: node }));
                }
                else {
                    hasDeprecation = true;
                    markDeprecated(program, type, {
                        message: messageStr,
                    });
                }
            }
        }
    }
    // the types here aren't ideal and could probably be refactored.
    function createAndFinishType(typeDef) {
        createType(typeDef);
        return finishType(typeDef);
    }
    /**
     * Given the own-properties of a type, returns a fully-initialized type.
     */
    function createType(typeDef) {
        stats.createdTypes++;
        Object.setPrototypeOf(typeDef, typePrototype);
        typeDef.isFinished = false;
        // If the type has an associated syntax node, check any directives that
        // might be attached.
        const createdType = typeDef;
        createdType.entityKind = "Type";
        if (createdType.node) {
            checkDirectives(createdType.node, createdType);
        }
        return createdType;
    }
    function finishType(typeDef) {
        stats.finishedTypes++;
        return finishTypeForProgramAndChecker(program, typePrototype, typeDef);
    }
    function getLiteralType(node) {
        return createLiteralType(node.value, node);
    }
    function getMergedSymbol(sym) {
        // if (!sym) return sym;
        // return mergedSymbols.get(sym) || sym;
        return resolver.getMergedSymbol(sym);
    }
    function createGlobalNamespaceType() {
        const sym = resolver.symbols.global;
        const type = createType({
            kind: "Namespace",
            name: "",
            node: getGlobalNamespaceNode(),
            models: new Map(),
            scalars: new Map(),
            operations: new Map(),
            namespaces: new Map(),
            interfaces: new Map(),
            unions: new Map(),
            enums: new Map(),
            decoratorDeclarations: new Map(),
            functionDeclarations: new Map(),
            decorators: [],
        });
        getSymbolLinks(sym).type = type;
        type.decorators = checkAugmentDecorators(sym, type, undefined);
        return finishType(type);
    }
    function initializeClone(type, additionalProps) {
        let clone;
        switch (type.kind) {
            case "Model":
                const newModel = createType({
                    ...type,
                    decorators: [...type.decorators],
                    properties: undefined,
                    ...additionalProps,
                });
                if (!("properties" in additionalProps)) {
                    newModel.properties = createRekeyableMap(Array.from(type.properties.entries()).map(([key, prop]) => [
                        key,
                        cloneType(prop, { model: newModel }),
                    ]));
                }
                clone = newModel;
                break;
            case "Union":
                const newUnion = createType({
                    ...type,
                    decorators: [...type.decorators],
                    variants: undefined,
                    get options() {
                        return Array.from(this.variants.values()).map((v) => v.type);
                    },
                    ...additionalProps,
                });
                if (!("variants" in additionalProps)) {
                    newUnion.variants = createRekeyableMap(Array.from(type.variants.entries()).map(([key, prop]) => [
                        key,
                        cloneType(prop, { union: newUnion }),
                    ]));
                }
                clone = newUnion;
                break;
            case "Interface":
                const newInterface = createType({
                    ...type,
                    decorators: [...type.decorators],
                    operations: undefined,
                    ...additionalProps,
                });
                if (!("operations" in additionalProps)) {
                    newInterface.operations = createRekeyableMap(Array.from(type.operations.entries()).map(([key, prop]) => [
                        key,
                        cloneType(prop, { interface: newInterface }),
                    ]));
                }
                clone = newInterface;
                break;
            case "Enum":
                const newEnum = createType({
                    ...type,
                    decorators: [...type.decorators],
                    members: undefined,
                    ...additionalProps,
                });
                if (!("members" in additionalProps)) {
                    newEnum.members = createRekeyableMap(Array.from(type.members.entries()).map(([key, prop]) => [
                        key,
                        cloneType(prop, { enum: newEnum }),
                    ]));
                }
                clone = newEnum;
                break;
            default:
                clone = createType({
                    ...type,
                    ...("decorators" in type ? { decorators: [...type.decorators] } : {}),
                    ...additionalProps,
                });
                break;
        }
        return clone;
    }
    /**
     * Clone a type, resulting in an identical type with all the same decorators
     * applied. Decorators are re-run on the clone to achieve this.
     *
     * Care is taken to clone nested data structures that are part of the type.
     * Any type with e.g. a map or an array property must recreate the map or array
     * so that clones don't share the same object.
     *
     * For types which have sub-types that are part of it, e.g. enums with members,
     * unions with variants, or models with properties, the sub-types are cloned
     * as well.
     *
     * If the entire type graph needs to be cloned, then cloneType must be called
     * recursively by the caller.
     */
    function cloneType(type, additionalProps = {}) {
        let clone = initializeClone(type, additionalProps);
        if (type.isFinished) {
            clone = finishType(clone);
        }
        compilerAssert(clone.kind === type.kind, "cloneType must not change type kind");
        return clone;
    }
    /**
     * Clone a type linking to the given symbol.
     * @param sym Symbol which to associate the clone
     * @param type Type to clone
     * @param additionalProps Additional properties to set/override on the clone
     * @returns cloned type
     */
    function cloneTypeForSymbol(sym, type, additionalProps = {}) {
        let clone = initializeClone(type, additionalProps);
        if ("decorators" in clone) {
            const docComment = docFromCommentForSym.get(sym);
            if (docComment) {
                clone.decorators.push(createDocFromCommentDecorator("self", docComment));
            }
            for (const dec of checkAugmentDecorators(sym, clone, undefined)) {
                clone.decorators.push(dec);
            }
        }
        if (type.isFinished) {
            clone = finishType(clone);
        }
        compilerAssert(clone.kind === type.kind, "cloneType must not change type kind");
        return clone;
    }
    function createLiteralType(value, node) {
        if (program.literalTypes.has(value)) {
            return program.literalTypes.get(value);
        }
        let type;
        switch (typeof value) {
            case "string":
                type = createType({ kind: "String", value });
                break;
            case "boolean":
                type = createType({ kind: "Boolean", value });
                break;
            case "number":
                let valueAsString;
                if (node) {
                    compilerAssert(node.kind === SyntaxKind.NumericLiteral, "Must pass numeric literal node or undefined when creating a numeric literal type");
                    valueAsString = node.valueAsString;
                }
                else {
                    valueAsString = String(value);
                }
                type = createType({
                    kind: "Number",
                    value,
                    valueAsString,
                    numericValue: Numeric(valueAsString),
                });
                break;
        }
        program.literalTypes.set(value, type);
        return type;
    }
    /**
     * Check if the source type can be assigned to the target type and emit diagnostics
     * @param source Type of a value
     * @param constraint
     * @param diagnosticTarget Target for the diagnostic, unless something better can be inferred.
     */
    function checkTypeOfValueMatchConstraint(source, constraint, diagnosticTarget) {
        const [related, diagnostics] = relation.isTypeAssignableTo(source, constraint.type, diagnosticTarget);
        if (!related) {
            if (constraint.kind === "argument") {
                reportCheckerDiagnostic(createDiagnostic({
                    code: "invalid-argument",
                    format: {
                        value: getEntityName(source),
                        expected: getEntityName(constraint.type),
                    },
                    target: diagnosticTarget,
                }));
            }
            else {
                reportCheckerDiagnostics(diagnostics);
            }
        }
        return related;
    }
    /**
     * Check if the source type can be assigned to the target type and emit diagnostics
     * @param source Source type
     * @param target Target type
     * @param diagnosticTarget Target for the diagnostic, unless something better can be inferred.
     */
    function checkTypeAssignable(source, target, diagnosticTarget) {
        const [related, diagnostics] = relation.isTypeAssignableTo(source, target, diagnosticTarget);
        if (!related) {
            reportCheckerDiagnostics(diagnostics);
        }
        return related;
    }
    function checkValueOfType(source, target, diagnosticTarget) {
        const [related, diagnostics] = relation.isValueOfType(source, target, diagnosticTarget);
        if (!related) {
            reportCheckerDiagnostics(diagnostics);
        }
        return related;
    }
    function isStdType(type, stdType) {
        if ((type.kind !== "Model" && type.kind !== "Scalar") ||
            type.namespace === undefined ||
            !isTypeSpecNamespace(type.namespace))
            return false;
        if (type.kind === "Scalar")
            return stdType === undefined || stdType === type.name;
        if (stdType === "Array" && type === stdTypes["Array"])
            return true;
        if (stdType === "Record" && type === stdTypes["Record"])
            return true;
        if (type.kind === "Model")
            return stdType === undefined || stdType === type.name;
        return false;
    }
    function getValueExactType(value) {
        return valueExactTypes.get(value);
    }
}
/**
 * Find all named models that could have been the source of the given
 * property. This includes the named parents of all property sources in a
 * chain.
 */
function getNamedSourceModels(property) {
    if (!property.sourceProperty) {
        return undefined;
    }
    const set = new Set();
    for (let p = property; p; p = p.sourceProperty) {
        if (p.model?.name) {
            set.add(p.model);
        }
    }
    return set;
}
/**
 * Find derived types of `models` in `possiblyDerivedModels` and add them to
 * `models`.
 */
function addDerivedModels(models, possiblyDerivedModels) {
    for (const element of possiblyDerivedModels) {
        if (!models.has(element)) {
            for (let t = element.baseModel; t; t = t.baseModel) {
                if (models.has(t)) {
                    models.add(element);
                    break;
                }
            }
        }
    }
}
function createTypeMapper(parameters, args, source, parentMapper) {
    const map = new Map(parentMapper?.map ?? []);
    for (const [index, param] of parameters.entries()) {
        map.set(param, args[index]);
    }
    return {
        partial: false,
        args: [...(parentMapper?.args ?? []), ...args],
        getMappedType: (type) => {
            return map.get(type) ?? type;
        },
        source,
        map,
    };
}
/**
 * If the input is anonymous (or the provided filter removes properties)
 * and there exists a named model with the same set of properties
 * (ignoring filtered properties), then return that named model.
 * Otherwise, return the input unchanged.
 *
 * This can be used by emitters to find a better name for a set of
 * properties after filtering. For example, given `{ @metadata prop:
 * string} & SomeName`, and an emitter that wishes to discard properties
 * marked with `@metadata`, the emitter can use this to recover that the
 * best name for the remaining properties is `SomeName`.
 *
 * @param model The input model
 * @param filter An optional filter to apply to the input model's
 * properties.
 */
export function getEffectiveModelType(program, model, filter) {
    if (filter) {
        model = filterModelProperties(program, model, filter);
    }
    if (model.name) {
        // named model
        return model;
    }
    // We would need to change the algorithm if this doesn't hold. We
    // assume model has no inherited properties below.
    compilerAssert(!model.baseModel, "Anonymous model with base model.");
    if (model.properties.size === 0) {
        // empty model
        return model;
    }
    // Find the candidate set of named model types that could have been the
    // source of every property in the model.
    let candidates;
    for (const property of model.properties.values()) {
        const sources = getNamedSourceModels(property);
        if (!sources) {
            // unsourced property: no possible match
            return model;
        }
        if (!candidates) {
            // first sourced property: initialize candidates to its sources
            candidates = sources;
            continue;
        }
        // Add derived sources as we encounter them. If a model is sourced from
        // a base property, then it can also be sourced from a derived model.
        //
        // (Unless it is overridden, but then the presence of the overridden
        // property will still cause the the base model to be excluded from the
        // candidates.)
        //
        // Note: We depend on the order of that spread and intersect source
        // properties here, which is that we see properties sourced from derived
        // types before properties sourced from their base types.
        addDerivedModels(sources, candidates);
        // remove candidates that are not common to this property.
        for (const candidate of candidates) {
            if (!sources.has(candidate)) {
                candidates.delete(candidate);
            }
        }
    }
    // Search for a candidate that has no additional properties (ignoring
    // filtered properties). If so, it is effectively the same type as the
    // input model. Consider a candidate that meets this test without
    // ignoring filtering as a better match than one that requires filtering
    // to meet this test.
    let match;
    for (const candidate of candidates ?? []) {
        if (model.properties.size === countPropertiesInherited(candidate)) {
            match = candidate;
            break; // exact match
        }
        if (filter && !match && model.properties.size === countPropertiesInherited(candidate, filter)) {
            match = candidate;
            continue; // match with filter: keep searching for exact match
        }
    }
    return match ?? model;
}
/**
 * Applies a filter to the properties of a given type. If no properties
 * are filtered out, then return the input unchanged. Otherwise, return
 * a new anonymous model with only the filtered properties.
 *
 * @param model The input model to filter.
 * @param filter The filter to apply. Properties are kept when this returns true.
 */
export function filterModelProperties(program, model, filter) {
    let filtered = false;
    for (const property of walkPropertiesInherited(model)) {
        if (!filter(property)) {
            filtered = true;
            break;
        }
    }
    if (!filtered) {
        return model;
    }
    const realm = Realm.realmForType.get(model);
    const typekit = realm ? $(realm) : $(program);
    const newModel = typekit.model.create({
        name: "",
        indexer: undefined,
        properties: {},
        decorators: [],
        derivedModels: [],
        sourceModels: [{ usage: "spread", model }],
    });
    for (const property of walkPropertiesInherited(model)) {
        if (filter(property)) {
            const newProperty = typekit.type.clone(property);
            Object.assign(newProperty, {
                sourceProperty: property,
                model: newModel,
            });
            newModel.properties.set(property.name, newProperty);
            typekit.type.finishType(newProperty);
        }
    }
    return finishTypeForProgram(program, newModel);
}
/**
 * Gets the property from the nearest base type that is overridden by the
 * given property, if any.
 */
export function getOverriddenProperty(property) {
    compilerAssert(property.model, "Parent model must be set before overridden property can be found.");
    for (let current = property.model.baseModel; current; current = current.baseModel) {
        const overridden = current.properties.get(property.name);
        if (overridden) {
            return overridden;
        }
    }
    return undefined;
}
/**
 * Enumerates the properties declared by model or inherited from its base.
 *
 * Properties declared by more derived types are enumerated before properties
 * of less derived types.
 *
 * Properties that are overridden are not enumerated.
 */
export function* walkPropertiesInherited(model) {
    const returned = new Set();
    for (let current = model; current; current = current.baseModel) {
        for (const property of current.properties.values()) {
            if (returned.has(property.name)) {
                // skip properties that have been overridden
                continue;
            }
            returned.add(property.name);
            yield property;
        }
    }
}
function countPropertiesInherited(model, filter) {
    let count = 0;
    for (const property of walkPropertiesInherited(model)) {
        if (!filter || filter(property)) {
            count++;
        }
    }
    return count;
}
export function finishTypeForProgram(program, typeDef) {
    return finishTypeForProgramAndChecker(program, program.checker.typePrototype, typeDef);
}
function linkMapper(typeDef, mapper) {
    if (mapper) {
        compilerAssert(!typeDef.templateMapper, "Mapper provided but template arguments already set.");
        typeDef.templateMapper = mapper;
    }
}
function extractMainDoc(type) {
    if (type.node?.docs === undefined) {
        return undefined;
    }
    let mainDoc = "";
    for (const doc of type.node.docs) {
        mainDoc += getDocContent(doc.content);
    }
    const trimmed = mainDoc.trim();
    return trimmed === "" ? undefined : trimmed;
}
function extractReturnsDocs(type) {
    const result = {
        returns: undefined,
        errors: undefined,
    };
    if (type.node?.docs === undefined) {
        return result;
    }
    for (const doc of type.node.docs) {
        for (const tag of doc.tags) {
            if (tag.kind === SyntaxKind.DocReturnsTag) {
                result.returns = getDocContent(tag.content);
            }
            if (tag.kind === SyntaxKind.DocErrorsTag) {
                result.errors = getDocContent(tag.content);
            }
        }
    }
    return result;
}
function extractParamDocs(node) {
    if (node.docs === undefined) {
        return new Map();
    }
    const paramDocs = new Map();
    for (const doc of node.docs) {
        for (const tag of doc.tags) {
            if (tag.kind === SyntaxKind.DocParamTag) {
                paramDocs.set(tag.paramName.sv, getDocContent(tag.content));
            }
        }
    }
    return paramDocs;
}
function extractPropDocs(node) {
    if (node.docs === undefined) {
        return new Map();
    }
    const propDocs = new Map();
    for (const doc of node.docs) {
        for (const tag of doc.tags) {
            if (tag.kind === SyntaxKind.DocPropTag) {
                propDocs.set(tag.propName.sv, getDocContent(tag.content));
            }
        }
    }
    return propDocs;
}
function getDocContent(content) {
    const docs = [];
    for (const node of content) {
        compilerAssert(node.kind === SyntaxKind.DocText, "No other doc content node kinds exist yet. Update this code appropriately when more are added.");
        docs.push(node.text);
    }
    return docs.join("");
}
function finishTypeForProgramAndChecker(program, typePrototype, typeDef) {
    if ("decorators" in typeDef) {
        for (const decApp of typeDef.decorators) {
            applyDecoratorToType(program, decApp, typeDef);
        }
    }
    Object.setPrototypeOf(typeDef, typePrototype);
    typeDef.isFinished = true;
    return typeDef;
}
function reportDeprecation(program, target, message, reportFunc) {
    if (program.compilerOptions.ignoreDeprecated !== true) {
        reportFunc(createDiagnostic({
            code: "deprecated",
            format: {
                message,
            },
            target,
        }));
    }
}
function applyDecoratorToType(program, decApp, target) {
    compilerAssert("decorators" in target, "Cannot apply decorator to non-decoratable type", target);
    for (const arg of decApp.args) {
        if (isType(arg.value) && isErrorType(arg.value)) {
            // If one of the decorator argument is an error don't run it.
            return;
        }
    }
    // Is the decorator definition deprecated?
    if (decApp.definition) {
        const deprecation = getDeprecationDetails(program, decApp.definition);
        if (deprecation !== undefined) {
            reportDeprecation(program, decApp.node ?? target, deprecation.message, program.reportDiagnostic);
        }
    }
    // peel `fn` off to avoid setting `this`.
    try {
        const args = decApp.args.map((x) => x.jsValue);
        const fn = decApp.decorator;
        const context = createDecoratorContext(program, decApp);
        fn(context, target, ...args);
    }
    catch (error) {
        // do not fail the language server for exceptions in decorators
        if (program.compilerOptions.designTimeBuild) {
            program.reportDiagnostic(createDiagnostic({
                code: "decorator-fail",
                format: { decoratorName: decApp.decorator.name, error: error.stack },
                target: decApp.node ?? target,
            }));
        }
        else {
            throw error;
        }
    }
}
function createDecoratorContext(program, decApp) {
    function createPassThruContext(program, decApp) {
        return {
            program,
            decoratorTarget: decApp.node,
            getArgumentTarget: () => decApp.node,
            call: (decorator, target, ...args) => {
                return decorator(createPassThruContext(program, decApp), target, ...args);
            },
        };
    }
    return {
        program,
        decoratorTarget: decApp.node,
        getArgumentTarget: (index) => {
            return decApp.args[index]?.node;
        },
        call: (decorator, target, ...args) => {
            return decorator(createPassThruContext(program, decApp), target, ...args);
        },
    };
}
function isTemplatedNode(node) {
    return "templateParameters" in node && node.templateParameters.length > 0;
}
var ResolutionKind;
(function (ResolutionKind) {
    ResolutionKind[ResolutionKind["Value"] = 0] = "Value";
    ResolutionKind[ResolutionKind["Type"] = 1] = "Type";
    ResolutionKind[ResolutionKind["BaseType"] = 2] = "BaseType";
    ResolutionKind[ResolutionKind["Constraint"] = 3] = "Constraint";
})(ResolutionKind || (ResolutionKind = {}));
class PendingResolutions {
    #data = new Map();
    start(symId, kind) {
        let existing = this.#data.get(symId);
        if (existing === undefined) {
            existing = new Set();
            this.#data.set(symId, existing);
        }
        existing.add(kind);
    }
    has(symId, kind) {
        return this.#data.get(symId)?.has(kind) ?? false;
    }
    finish(symId, kind) {
        const existing = this.#data.get(symId);
        if (existing === undefined) {
            return;
        }
        existing?.delete(kind);
        if (existing.size === 0) {
            this.#data.delete(symId);
        }
    }
}
const defaultSymbolResolutionOptions = {
    resolveDecorators: false,
    checkTemplateTypes: true,
};
function printTypeReferenceNode(node) {
    switch (node.kind) {
        case SyntaxKind.MemberExpression:
            return `${printTypeReferenceNode(node.base)}.${printTypeReferenceNode(node.id)}`;
        case SyntaxKind.TypeReference:
            return printTypeReferenceNode(node.target);
        case SyntaxKind.Identifier:
            return node.sv;
    }
}
//# sourceMappingURL=checker.js.map