import { spawnSync } from "child_process";
import pc from "picocolors";
import { inspect } from "util";
import { logDiagnostics } from "../diagnostics.js";
import { ExternalError } from "../external-error.js";
import { createConsoleSink } from "../logger/console-sink.js";
import { createLogger } from "../logger/logger.js";
import { createTracer } from "../logger/tracer.js";
import { NodeHost } from "../node-host.js";
import { getBaseFileName } from "../path-utils.js";
// ENOENT checking and handles spaces poorly in some cases.
const isCmdOnWindows = ["code", "code-insiders", "npm"];
export function withCliHost(fn) {
    return withFailsafe((args) => {
        const host = createCLICompilerHost(args);
        return fn(host, args);
    });
}
function withFailsafe(fn) {
    return async (...args) => {
        try {
            return await fn(...args);
        }
        catch (error) {
            handleInternalCompilerError(error);
        }
    };
}
/**
 * Resolve Cli host automatically using cli args and handle diagnostics returned by the action.
 */
export function withCliHostAndDiagnostics(fn) {
    return withFailsafe(async (args) => {
        const host = createCLICompilerHost(args);
        const diagnostics = await fn(host, args);
        logDiagnostics(diagnostics, host.logSink);
        logDiagnosticCount(diagnostics);
        if (diagnostics.some((d) => d.severity === "error")) {
            process.exit(1);
        }
    });
}
export function createCLICompilerHost(options) {
    const logSink = createConsoleSink({
        pretty: options.pretty,
        pathRelativeTo: process.cwd(),
        trackAction: true,
    });
    const logger = createLogger({ sink: logSink });
    const tracer = createTracer(logger, {
        filter: options.trace ?? (options.debug ? ["*"] : undefined),
    });
    tracer.trace("cli.args", `CLI args: ${inspect(options, { depth: null })}`);
    return { ...NodeHost, logSink, logger, tracer, debug: options.debug ?? false };
}
export function run(host, command, commandArgs, options) {
    const logger = host.logger;
    if (options) {
        logger.trace(inspect(options, { depth: null }));
    }
    logger.trace(`> ${command} ${commandArgs.join(" ")}\n`);
    if (options?.extraEnv) {
        options.env = {
            ...(options?.env ?? process.env),
            ...options.extraEnv,
        };
    }
    const baseCommandName = getBaseFileName(command);
    if (process.platform === "win32" && isCmdOnWindows.includes(command)) {
        command += ".cmd";
    }
    const useShell = process.platform === "win32";
    const hasSpace = command.includes(" ");
    let commandToSpawn = command;
    if (useShell && hasSpace) {
        // for command having space (which should be a path), we need to wrap it into " for it to be executed properly in shell
        // but for short command like 'npm', we shouldn't wrap it which would trigger error
        commandToSpawn = `"${command}"`;
        logger.trace(`Command to spawn updated to: ${commandToSpawn}\n`);
    }
    const finalOptions = {
        encoding: "utf-8",
        stdio: "inherit",
        shell: useShell,
        ...(options ?? {}),
    };
    const proc = spawnSync(commandToSpawn, commandArgs, finalOptions);
    logger.trace(inspect(proc, { depth: null }));
    if (proc.error) {
        if (proc.error.code === "ENOENT" && !options?.allowNotFound) {
            logger.error(`error: Command '${baseCommandName}' not found.`);
            if (options?.debug && proc.error.stack) {
                logger.error(proc.error.stack);
            }
            process.exit(1);
        }
        else {
            throw proc.error;
        }
    }
    if (proc.status !== 0 && !options?.allowedExitCodes?.includes(proc.status ?? 0)) {
        logger.error(`error: Command '${baseCommandName} ${commandArgs.join(" ")}' failed with exit code ${proc.status}.`);
        process.exit(proc.status || 1);
    }
    return proc;
}
export function logDiagnosticCount(diagnostics) {
    if (diagnostics.length === 0) {
        return;
    }
    const errorCount = diagnostics.filter((x) => x.severity === "error").length;
    const warningCount = diagnostics.filter((x) => x.severity === "warning").length;
    const addSuffix = (count, suffix) => count > 1 ? `${count} ${suffix}s` : count === 1 ? `${count} ${suffix}` : undefined;
    const errorText = addSuffix(errorCount, "error");
    const warningText = addSuffix(warningCount, "warning");
    // eslint-disable-next-line no-console
    console.log(`\nFound ${[errorText, warningText].filter((x) => x !== undefined).join(", ")}.`);
}
export function logInternalCompilerError(error) {
    /* eslint-disable no-console */
    if (error instanceof ExternalError) {
        // ExternalError should already have all the relevant information needed when thrown.
        console.error(error.render(color));
    }
    else {
        console.error("Internal compiler error!");
        console.error("File issue at https://github.com/microsoft/typespec");
        console.error();
        console.error(error);
    }
    /* eslint-enable no-console */
}
/**
 * Handle an internal compiler error.
 *
 * NOTE: An expected error, like one thrown for bad input, shouldn't reach
 * here, but be handled somewhere else. If we reach here, it should be
 * considered a bug and therefore we should not suppress the stack trace as
 * that risks losing it in the case of a bug that does not repro easily.
 *
 * @param error error thrown
 */
export function handleInternalCompilerError(error) {
    logInternalCompilerError(error);
    process.exit(1);
}
function color(text, color) {
    return pc[color](text);
}
//# sourceMappingURL=utils.js.map